from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs
from ai.backend.manager.repositories.user.admin_repository import AdminUserRepository
from ai.backend.manager.repositories.user.repository import UserRepository


@dataclass
class UserRepositories:
    repository: UserRepository
    admin_repository: AdminUserRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = UserRepository(args.db)
        admin_repository = AdminUserRepository(args.db)

        return cls(
            repository=repository,
            admin_repository=admin_repository,
        )



from __future__ import annotations

from typing import Generic

from ai.backend.common.dependencies import DependencyProvider, ResourceT
from ai.backend.manager.config.bootstrap import BootstrapConfig


class BootstrapDependency(DependencyProvider[BootstrapConfig, ResourceT], Generic[ResourceT]):
    """Base class for bootstrap dependencies that only need BootstrapConfig."""

    pass



from __future__ import annotations

import logging
from typing import Sequence

import sqlalchemy as sa
from sqlalchemy.orm import foreign, relationship

from ai.backend.logging import BraceStyleAdapter

from .base import GUID, Base, IDColumn

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore

__all__: Sequence[str] = ("AssociationArtifactsStorageRow",)


def _get_association_artifact_join_cond():
    from .artifact_revision import ArtifactRevisionRow

    return ArtifactRevisionRow.id == foreign(AssociationArtifactsStorageRow.artifact_revision_id)


def _get_association_object_storage_join_cond():
    from .object_storage import ObjectStorageRow

    return ObjectStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


def _get_association_vfs_storage_join_cond():
    from .vfs_storage import VFSStorageRow

    return VFSStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


class AssociationArtifactsStorageRow(Base):
    """
    Association table for linking artifacts to storage namespace.
    """

    __tablename__ = "association_artifacts_storages"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("artifact_revision_id", name="uq_artifact_revision_id"),
    )

    id = IDColumn()
    artifact_revision_id = sa.Column(
        "artifact_revision_id",
        GUID,
        nullable=False,
    )
    storage_namespace_id = sa.Column(
        "storage_namespace_id",
        GUID,
        nullable=False,
    )
    storage_type = sa.Column("storage_type", sa.String, nullable=False)

    artifact_revision_row = relationship(
        "ArtifactRevisionRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_artifact_join_cond,
    )

    # only valid when storage_type is "object_storage"
    object_storage_row = relationship(
        "ObjectStorageRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_object_storage_join_cond,
        overlaps="vfs_storage_row",
    )

    # only valid when storage_type is "vfs"
    vfs_storage_row = relationship(
        "VFSStorageRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_vfs_storage_join_cond,
        overlaps="object_storage_row",
    )



"""Cache source for app_config repository operations."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any, Optional

from glide import Batch, ExpirySet, ExpiryType

from ai.backend.common.json import dump_json, load_json
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.clients.valkey_client.valkey_cache import ValkeyCache
from ai.backend.manager.data.app_config.types import MergedAppConfig
from ai.backend.manager.models.app_config import AppConfigScopeType
from ai.backend.manager.repositories.utils import suppress_with_log

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

count = 0


class AppConfigCacheSource:
    """
    Cache source for app config operations.
    Handles all Redis/Valkey cache operations for app configurations.
    """

    _valkey_cache: ValkeyCache
    _cache_ttl: int

    def __init__(self, valkey_cache: ValkeyCache, cache_ttl: int = 300) -> None:
        """
        Initialize cache source.

        Args:
            valkey_cache: Valkey cache client for caching
            cache_ttl: Cache TTL in seconds (default: 5 minutes)
        """
        self._valkey_cache = valkey_cache
        self._cache_ttl = cache_ttl

    def _get_merged_config_cache_key(self, user_id: str) -> str:
        """Generate cache key for merged config."""
        return f"app_config:merged:{user_id}"

    def _get_domain_users_set_key(self, domain_name: str) -> str:
        """Generate Redis Set key for tracking users in a domain."""
        return f"app_config:domain:{domain_name}:users"

    async def get_merged_config(self, user_id: str) -> Optional[Mapping[str, Any]]:
        """
        Get merged configuration from cache.

        Returns:
            Cached config if exists, None if not in cache
        """
        with suppress_with_log([Exception], "Failed to get merged config from cache"):
            cache_key = self._get_merged_config_cache_key(user_id)
            cached_value = await self._valkey_cache.client.get(cache_key)
            global count
            count += 1
            if cached_value:
                log.debug("Cache hit for merged config: {}, hit count: {}", user_id, count)
                return load_json(cached_value)
            log.debug("Cache miss for merged config: {}", user_id)
        return None

    async def set_merged_config(
        self,
        merged_config: MergedAppConfig,
    ) -> None:
        """
        Set merged configuration in cache.

        Also tracks this user_id in the domain's users Set for efficient invalidation.

        Args:
            merged_config: MergedAppConfig containing user_id, domain_name, and config
        """
        with suppress_with_log([Exception], "Failed to set merged config in cache"):
            # Use batch to set both the config and add user to domain set
            batch = Batch(is_atomic=False)

            # Cache the merged config
            cache_key = self._get_merged_config_cache_key(merged_config.user_id)
            batch.set(
                cache_key,
                dump_json(merged_config.merged_config),
                expiry=ExpirySet(ExpiryType.SEC, self._cache_ttl),
            )

            # Add user_id to domain's users Set
            domain_users_key = self._get_domain_users_set_key(merged_config.domain_name)
            batch.sadd(domain_users_key, [merged_config.user_id])
            batch.expire(domain_users_key, self._cache_ttl)

            await self._valkey_cache.client.exec(batch, raise_on_error=True)

            log.trace(
                "Cached merged config for user {} in domain {}",
                merged_config.user_id,
                merged_config.domain_name,
            )

    async def invalidate_config(
        self,
        scope_type: AppConfigScopeType,
        scope_id: str,
    ) -> None:
        """
        Invalidate cache for a specific config.

        When a config is updated:
        - Domain config change: Delete the domain users Set and use pattern matching to find all user configs
        - User config change: Invalidate only that user
        """
        with suppress_with_log([Exception], "Failed to invalidate config cache"):
            match scope_type:
                case AppConfigScopeType.DOMAIN:
                    domain_users_key = self._get_domain_users_set_key(scope_id)
                    user_ids = await self._valkey_cache.client.smembers(domain_users_key)
                    if not user_ids:
                        log.debug(
                            "No users found for domain: {}, skipping cache invalidation", scope_id
                        )
                        return
                    keys_to_delete: list[str | bytes] = [
                        self._get_merged_config_cache_key(user_id.decode()) for user_id in user_ids
                    ]
                    keys_to_delete.append(domain_users_key)
                    remove_count = await self._valkey_cache.client.delete(keys_to_delete)
                    log.debug(
                        "Invalidated {} merged config caches for domain: {}",
                        remove_count,
                        scope_id,
                    )
                case AppConfigScopeType.USER:
                    # For user-level config, only invalidate that user's merged config
                    cache_key = self._get_merged_config_cache_key(scope_id)
                    await self._valkey_cache.client.delete([cache_key])
                    log.debug("Invalidated merged config for user: {}", scope_id)

                case _:
                    # PROJECT or other future scope types
                    log.debug("No cache invalidation needed for scope type: {}", scope_type)

            log.trace("Invalidated config cache: {}:{}", scope_type.value, scope_id)



from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import Generator

from ai.backend.common.types import SessionId

from .types import ExecutionStep, PhaseDescriptor, StepDescriptor


class TransitionRecorder:
    """
    Records execution steps during scheduler lifecycle operations.

    Supports hierarchical phases and automatic success/failure recording via context managers.
    """

    def __init__(self, operation: str) -> None:
        self._operation = operation
        self._steps: dict[SessionId, list[ExecutionStep]] = defaultdict(list)
        self._phase_stack: dict[SessionId, list[str]] = defaultdict(list)

    @property
    def operation(self) -> str:
        """Get the operation name for this recorder."""
        return self._operation

    @contextmanager
    def phase(self, desc: PhaseDescriptor) -> Generator[None, None, None]:
        """
        Enter a phase (hierarchical grouping).

        Usage:
            with recorder.phase(PhaseDescriptor(session_id, "provisioner")):
                with recorder.phase(PhaseDescriptor(session_id, "validator")):
                    # steps here will have phase ["provisioner", "validator"]
        """
        self._phase_stack[desc.session_id].append(desc.name)
        try:
            yield
        finally:
            self._phase_stack[desc.session_id].pop()

    @contextmanager
    def step(self, desc: StepDescriptor) -> Generator[None, None, None]:
        """
        Execute a step with automatic success/failure recording.

        Records "started" on entry, "success" on normal exit, "failed" on exception.

        Usage:
            with recorder.step(StepDescriptor(session_id, "check_quota", "Quota validated")):
                await check_quota()
                # On success: records "success" with success_detail
                # On exception: records "failed" with exception message
        """
        phases = self._get_current_phases(desc.session_id)
        self._record(desc.session_id, phases, desc.name, "started", None)
        try:
            yield
            self._record(desc.session_id, phases, desc.name, "success", desc.success_detail)
        except Exception as e:
            self._record(desc.session_id, phases, desc.name, "failed", str(e))
            raise

    def _get_current_phases(self, session_id: SessionId) -> list[str]:
        """Get the current phase stack for a session."""
        return list(self._phase_stack[session_id])

    def _record(
        self,
        session_id: SessionId,
        phases: list[str],
        name: str,
        status: str,
        detail: str | None,
    ) -> None:
        """Internal method to record an execution step."""
        self._steps[session_id].append(
            ExecutionStep(
                phases=phases,
                name=name,
                status=status,
                timestamp=datetime.now(timezone.utc),
                detail=detail,
            )
        )

    def get_steps(self, session_id: SessionId) -> list[ExecutionStep]:
        """Get all recorded steps for a specific session."""
        return list(self._steps[session_id])

    def get_all_steps(self) -> dict[SessionId, list[ExecutionStep]]:
        """Get all recorded steps for all sessions."""
        return dict(self._steps)



import uuid

import sqlalchemy as sa

from ai.backend.manager.data.vfs_storage.creator import VFSStorageCreator
from ai.backend.manager.data.vfs_storage.modifier import VFSStorageModifier
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.errors.vfs_storage import (
    VFSStorageNotFoundError,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.models.vfs_storage import VFSStorageRow


class VFSStorageDBSource:
    """Database source for VFS storage operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def get_by_name(self, storage_name: str) -> VFSStorageData:
        """
        Get an existing VFS storage configuration from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(VFSStorageRow).where(VFSStorageRow.name == storage_name)
            result = await db_session.execute(query)
            row: VFSStorageRow = result.scalar_one_or_none()
            if row is None:
                raise VFSStorageNotFoundError(f"VFS storage with name {storage_name} not found.")
            return row.to_dataclass()

    async def get_by_id(self, storage_id: uuid.UUID) -> VFSStorageData:
        """
        Get an existing VFS storage configuration from the database by ID.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(VFSStorageRow).where(VFSStorageRow.id == storage_id)
            result = await db_session.execute(query)
            row: VFSStorageRow = result.scalar_one_or_none()
            if row is None:
                raise VFSStorageNotFoundError(f"VFS storage with ID {storage_id} not found.")
            return row.to_dataclass()

    async def create(self, creator: VFSStorageCreator) -> VFSStorageData:
        """
        Create a new VFS storage configuration in the database.
        """
        async with self._db.begin_session() as db_session:
            vfs_storage_data = creator.fields_to_store()
            vfs_storage_row = VFSStorageRow(**vfs_storage_data)
            db_session.add(vfs_storage_row)
            await db_session.flush()
            await db_session.refresh(vfs_storage_row)
            return vfs_storage_row.to_dataclass()

    async def update(self, storage_id: uuid.UUID, modifier: VFSStorageModifier) -> VFSStorageData:
        """
        Update an existing VFS storage configuration in the database.
        """
        async with self._db.begin_session() as db_session:
            data = modifier.fields_to_update()
            update_stmt = (
                sa.update(VFSStorageRow)
                .where(VFSStorageRow.id == storage_id)
                .values(**data)
                .returning(*sa.select(VFSStorageRow).selected_columns)
            )
            stmt = sa.select(VFSStorageRow).from_statement(update_stmt)
            row: VFSStorageRow = (await db_session.execute(stmt)).scalars().one()

            return row.to_dataclass()

    async def delete(self, storage_id: uuid.UUID) -> uuid.UUID:
        """
        Delete an existing VFS storage configuration from the database.
        """
        async with self._db.begin_session() as db_session:
            delete_query = (
                sa.delete(VFSStorageRow)
                .where(VFSStorageRow.id == storage_id)
                .returning(VFSStorageRow.id)
            )
            result = await db_session.execute(delete_query)
            deleted_id = result.scalar()
            return deleted_id

    async def list_vfs_storages(self) -> list[VFSStorageData]:
        """
        List all VFS storage configurations from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(VFSStorageRow)
            result = await db_session.execute(query)
            rows: list[VFSStorageRow] = result.scalars().all()
            return [row.to_dataclass() for row in rows]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.types import ModelReplicaData
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class BatchLoadReplicasByRevisionIdsAction(DeploymentBaseAction):
    revision_ids: list[UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "batch_load_replicas_by_revision_ids"


@dataclass
class BatchLoadReplicasByRevisionIdsActionResult(BaseActionResult):
    data: dict[UUID, list[ModelReplicaData]]

    @override
    def entity_id(self) -> Optional[str]:
        return None  # This is a list operation for replicas



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.notification import (
    NotificationChannelCreator,
    NotificationChannelData,
)

from .base import NotificationAction


@dataclass
class CreateChannelAction(NotificationAction):
    """Action to create a notification channel."""

    creator: NotificationChannelCreator

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create_channel"

    @override
    def entity_id(self) -> Optional[str]:
        return self.creator.name


@dataclass
class CreateChannelActionResult(BaseActionResult):
    """Result of creating a notification channel."""

    channel_data: NotificationChannelData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_data.id)



import logging
from datetime import datetime

from ai.backend.common.events.event_types.image.anycast import (
    ImagePullFailedEvent,
    ImagePullFinishedEvent,
    ImagePullStartedEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator

from ...models.utils import (
    ExtendedAsyncSAEngine,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ImageEventHandler:
    def __init__(
        self,
        registry: AgentRegistry,
        db: ExtendedAsyncSAEngine,
        use_sokovan: bool,
        schedule_coordinator: ScheduleCoordinator,
    ) -> None:
        self._registry = registry
        self._db = db
        self._use_sokovan = use_sokovan
        self._schedule_coordinator = schedule_coordinator

    async def handle_image_pull_started(
        self,
        context: None,
        agent_id: AgentId,
        ev: ImagePullStartedEvent,
    ) -> None:
        dt = datetime.fromtimestamp(ev.timestamp)
        log.debug("handle_image_pull_started: ag:{} img:{}, start_dt:{}", ev.agent_id, ev.image, dt)

        if self._use_sokovan:
            # Use new Sokovan logic for kernel state management
            await self._schedule_coordinator.update_kernels_to_pulling_for_image(
                ev.agent_id, ev.image, ev.image_ref.canonical if ev.image_ref else None
            )
        else:
            # Use legacy registry logic
            async with self._db.connect() as db_conn:
                await self._registry.mark_image_pull_started(
                    ev.agent_id, ev.image, ev.image_ref, db_conn=db_conn
                )

    async def handle_image_pull_finished(
        self, context: None, agent_id: AgentId, ev: ImagePullFinishedEvent
    ) -> None:
        dt = datetime.fromtimestamp(ev.timestamp)
        log.debug("handle_image_pull_finished: ag:{} img:{}, end_dt:{}", ev.agent_id, ev.image, dt)

        if self._use_sokovan:
            # Use new Sokovan logic for kernel state management
            await self._schedule_coordinator.update_kernels_to_prepared_for_image(
                ev.agent_id, ev.image, ev.image_ref.canonical if ev.image_ref else None
            )
        else:
            # Use legacy registry logic
            async with self._db.connect() as db_conn:
                await self._registry.mark_image_pull_finished(
                    ev.agent_id, ev.image, ev.image_ref, db_conn=db_conn
                )

    async def handle_image_pull_failed(
        self,
        context: None,
        agent_id: AgentId,
        ev: ImagePullFailedEvent,
    ) -> None:
        log.warning("handle_image_pull_failed: ag:{} img:{}, msg:{}", ev.agent_id, ev.image, ev.msg)

        if self._use_sokovan:
            # Use new Sokovan logic for kernel state management
            await self._schedule_coordinator.cancel_kernels_for_failed_image(
                ev.agent_id, ev.image, ev.msg, ev.image_ref.canonical if ev.image_ref else None
            )
        else:
            # Use legacy registry logic
            async with self._db.connect() as db_conn:
                await self._registry.handle_image_pull_failed(
                    ev.agent_id, ev.image, ev.msg, ev.image_ref, db_conn=db_conn
                )



"""add role tables

Revision ID: 60bcbf00a96e
Revises: bf39b34717d4
Create Date: 2025-07-08 17:07:24.636221

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "60bcbf00a96e"
down_revision = "bf39b34717d4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "resource_permissions",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("entity_type", sa.String(length=32), nullable=False),
        sa.Column("entity_id", sa.String(length=64), nullable=False),
        sa.Column("operation", sa.String(length=32), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_resource_permissions")),
    )
    op.create_table(
        "role_permissions",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("entity_type", sa.String(length=32), nullable=False),
        sa.Column("operation", sa.String(length=32), nullable=False),
        sa.Column("scope_type", sa.String(length=32), nullable=False),
        sa.Column("scope_id", sa.String(length=64), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_role_permissions")),
    )
    op.create_table(
        "roles",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("state", sa.VARCHAR(length=16), server_default="active", nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("deleted_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_roles")),
    )
    op.create_table(
        "user_roles",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("user_id", GUID(), nullable=False),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("state", sa.VARCHAR(length=16), server_default="active", nullable=False),
        sa.Column("granted_by", GUID(), nullable=True),
        sa.Column(
            "granted_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("expires_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("deleted_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_user_roles")),
    )


def downgrade() -> None:
    op.drop_table("user_roles")
    op.drop_table("roles")
    op.drop_table("role_permissions")
    op.drop_table("resource_permissions")



from typing import Optional

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.data.image.types import ImageStatus
from ai.backend.manager.errors.image import ContainerRegistryNotFound
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.image import ImageRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, SASession

container_registry_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.CONTAINER_REGISTRY_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ContainerRegistryRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @container_registry_repository_resilience.apply()
    async def get_by_registry_and_project(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        async with self._db.begin_readonly_session() as session:
            result = await self._get_by_registry_and_project(session, registry_name, project)
            if not result:
                raise ContainerRegistryNotFound()
            return result

    @container_registry_repository_resilience.apply()
    async def get_by_registry_name(self, registry_name: str) -> list[ContainerRegistryData]:
        async with self._db.begin_readonly_session() as session:
            return await self._get_by_registry_name(session, registry_name)

    @container_registry_repository_resilience.apply()
    async def get_all(self) -> list[ContainerRegistryData]:
        async with self._db.begin_readonly_session() as session:
            return await self._get_all(session)

    @container_registry_repository_resilience.apply()
    async def clear_images(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        async with self._db.begin_session() as session:
            # Clear images
            update_stmt = (
                sa.update(ImageRow)
                .where(ImageRow.registry == registry_name)
                .where(ImageRow.status != ImageStatus.DELETED)
                .values(status=ImageStatus.DELETED)
            )
            if project:
                update_stmt = update_stmt.where(ImageRow.project == project)

            await session.execute(update_stmt)

            # Return registry data
            result = await self._get_by_registry_and_project(session, registry_name, project)
            if not result:
                raise ContainerRegistryNotFound()
            return result

    @container_registry_repository_resilience.apply()
    async def get_known_registries(self) -> dict[str, str]:
        async with self._db.begin_readonly_session() as session:
            known_registries_map = await ContainerRegistryRow.get_known_container_registries(
                session
            )

            known_registries = {}
            for project, registries in known_registries_map.items():
                for registry_name, url in registries.items():
                    if project not in known_registries:
                        known_registries[f"{project}/{registry_name}"] = url.human_repr()

            return known_registries

    @container_registry_repository_resilience.apply()
    async def get_registry_row_for_scanner(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryRow:
        """
        Get the raw ContainerRegistryRow object needed for container registry scanner.
        Raises ContainerRegistryNotFound if registry is not found.
        """
        async with self._db.begin_readonly_session() as session:
            stmt = sa.select(ContainerRegistryRow).where(
                ContainerRegistryRow.registry_name == registry_name,
            )
            if project:
                stmt = stmt.where(ContainerRegistryRow.project == project)

            row: Optional[ContainerRegistryRow] = await session.scalar(stmt)
            if not row:
                raise ContainerRegistryNotFound()
            return row

    async def _get_by_registry_and_project(
        self,
        session: SASession,
        registry_name: str,
        project: Optional[str] = None,
    ) -> Optional[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow).where(
            ContainerRegistryRow.registry_name == registry_name,
        )
        if project:
            stmt = stmt.where(ContainerRegistryRow.project == project)

        row: Optional[ContainerRegistryRow] = await session.scalar(stmt)
        return row.to_dataclass() if row else None

    async def _get_by_registry_name(
        self,
        session: SASession,
        registry_name: str,
    ) -> list[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow).where(
            ContainerRegistryRow.registry_name == registry_name
        )
        result = await session.execute(stmt)
        rows: list[ContainerRegistryRow] = result.scalars().all()
        return [row.to_dataclass() for row in rows]

    async def _get_all(self, session: SASession) -> list[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow)
        result = await session.execute(stmt)
        rows: list[ContainerRegistryRow] = result.scalars().all()
        return [row.to_dataclass() for row in rows]



from .container_registry.harbor import (
    AbstractPerProjectContainerRegistryQuotaService,
)


class ServicesContext:
    """
    In the API layer, requests are processed through the ServicesContext and
    its subordinate layers, including the DB, Client, and Repository layers.
    Each layer separates the responsibilities specific to its respective level.
    """

    per_project_container_registries_quota: AbstractPerProjectContainerRegistryQuotaService

    def __init__(
        self,
        per_project_container_registries_quota: AbstractPerProjectContainerRegistryQuotaService,
    ) -> None:
        self.per_project_container_registries_quota = per_project_container_registries_quota



from typing import Optional

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from .repository import ContainerRegistryRepository

container_registry_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.CONTAINER_REGISTRY_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class AdminContainerRegistryRepository:
    _repository: ContainerRegistryRepository

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._repository = ContainerRegistryRepository(db)

    @container_registry_repository_resilience.apply()
    async def clear_images_force(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        """
        Forcefully clear images from a container registry without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.clear_images(registry_name, project)

    @container_registry_repository_resilience.apply()
    async def get_by_registry_and_project_force(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        """
        Forcefully get container registry by name and project without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_by_registry_and_project(registry_name, project)

    @container_registry_repository_resilience.apply()
    async def get_by_registry_name_force(self, registry_name: str) -> list[ContainerRegistryData]:
        """
        Forcefully get container registries by name without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_by_registry_name(registry_name)

    @container_registry_repository_resilience.apply()
    async def get_all_force(self) -> list[ContainerRegistryData]:
        """
        Forcefully get all container registries without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_all()



import logging

from sqlalchemy import exc as sa_exc

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.domain.admin_repository import AdminDomainRepository
from ai.backend.manager.repositories.domain.repository import DomainRepository
from ai.backend.manager.services.domain.actions.create_domain import (
    CreateDomainAction,
    CreateDomainActionResult,
)
from ai.backend.manager.services.domain.actions.create_domain_node import (
    CreateDomainNodeAction,
    CreateDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.delete_domain import (
    DeleteDomainAction,
    DeleteDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain import (
    ModifyDomainAction,
    ModifyDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain_node import (
    ModifyDomainNodeAction,
    ModifyDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.purge_domain import (
    PurgeDomainAction,
    PurgeDomainActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class DomainService:
    _repository: DomainRepository
    _admin_repository: AdminDomainRepository

    def __init__(
        self,
        repository: DomainRepository,
        admin_repository: AdminDomainRepository,
    ) -> None:
        self._repository = repository
        self._admin_repository = admin_repository

    async def create_domain(self, action: CreateDomainAction) -> CreateDomainActionResult:
        try:
            if action.user_info.role == UserRole.SUPERADMIN:
                domain_data = await self._admin_repository.create_domain_force(action.creator)
            else:
                domain_data = await self._repository.create_domain_validated(action.creator)
            return CreateDomainActionResult(
                domain_data=domain_data,
                success=True,
                description="domain creation succeed",
            )
        except Exception as e:
            return CreateDomainActionResult(
                domain_data=None,
                success=False,
                description=f"domain creation failed: {str(e)}",
            )

    async def modify_domain(self, action: ModifyDomainAction) -> ModifyDomainActionResult:
        try:
            if action.user_info.role == UserRole.SUPERADMIN:
                domain_data = await self._admin_repository.modify_domain_force(
                    action.domain_name, action.modifier
                )
            else:
                domain_data = await self._repository.modify_domain_validated(
                    action.domain_name, action.modifier
                )
            if domain_data:
                return ModifyDomainActionResult(
                    domain_data=domain_data,
                    success=True,
                    description="domain modification succeed",
                )
            else:
                return ModifyDomainActionResult(
                    domain_data=None,
                    success=False,
                    description=f"no matching {action.domain_name}",
                )
        except Exception as e:
            return ModifyDomainActionResult(
                domain_data=None,
                success=False,
                description=f"domain modification failed: {str(e)}",
            )

    async def delete_domain(self, action: DeleteDomainAction) -> DeleteDomainActionResult:
        try:
            if action.user_info.role == UserRole.SUPERADMIN:
                success = await self._admin_repository.soft_delete_domain_force(action.name)
            else:
                success = await self._repository.soft_delete_domain_validated(action.name)
            if success:
                return DeleteDomainActionResult(
                    success=True,
                    description=f"domain {action.name} deleted successfully",
                )
            else:
                return DeleteDomainActionResult(
                    success=False, description=f"no matching {action.name}"
                )
        except Exception as e:
            return DeleteDomainActionResult(
                success=False, description=f"domain deletion failed: {str(e)}"
            )

    async def purge_domain(self, action: PurgeDomainAction) -> PurgeDomainActionResult:
        name = action.name

        if action.user_info.role == UserRole.SUPERADMIN:
            await self._admin_repository.purge_domain_force(name)
        else:
            await self._repository.purge_domain_validated(name)
        return PurgeDomainActionResult(
            success=True, description=f"domain {name} purged successfully"
        )

    async def create_domain_node(
        self, action: CreateDomainNodeAction
    ) -> CreateDomainNodeActionResult:
        scaling_groups = action.scaling_groups

        try:
            if action.user_info.role == UserRole.SUPERADMIN:
                domain_data = (
                    await self._admin_repository.create_domain_node_with_permissions_force(
                        action.creator, action.user_info, scaling_groups
                    )
                )
            else:
                domain_data = await self._repository.create_domain_node_with_permissions(
                    action.creator, action.user_info, scaling_groups
                )
        except sa_exc.IntegrityError as e:
            raise ValueError(
                f"Cannot create the domain with given arguments. (arg:{action}, e:{str(e)})"
            )

        return CreateDomainNodeActionResult(
            domain_data=domain_data,
            success=True,
            description=f"domain {action.creator.name} created",
        )

    async def modify_domain_node(
        self, action: ModifyDomainNodeAction
    ) -> ModifyDomainNodeActionResult:
        domain_name = action.name

        if action.sgroups_to_add is not None and action.sgroups_to_remove is not None:
            if union := action.sgroups_to_add | action.sgroups_to_remove:
                raise ValueError(
                    "Should be no scaling group names included in both `sgroups_to_add` and `sgroups_to_remove` "
                    f"(sg:{union})."
                )

        try:
            if action.user_info.role == UserRole.SUPERADMIN:
                domain_data = (
                    await self._admin_repository.modify_domain_node_with_permissions_force(
                        domain_name,
                        action.modifier.fields_to_update(),
                        action.user_info,
                        action.sgroups_to_add,
                        action.sgroups_to_remove,
                    )
                )
            else:
                domain_data = await self._repository.modify_domain_node_with_permissions(
                    domain_name,
                    action.modifier.fields_to_update(),
                    action.user_info,
                    action.sgroups_to_add,
                    action.sgroups_to_remove,
                )
        except sa_exc.IntegrityError as e:
            raise ValueError(
                f"Cannot modify the domain with given arguments. (arg:{action}, e:{str(e)})"
            )
        if domain_data is None:
            raise ValueError(f"Domain not found (id:{domain_name})")

        return ModifyDomainNodeActionResult(
            domain_data=domain_data,
            success=True,
            description=f"domain {domain_name} modified",
        )



from __future__ import annotations

import logging
import uuid
from collections.abc import Mapping, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Self,
)

import graphene
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.orm import contains_eager

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    HardwareMetadata,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.bgtask.tasks.rescan_gpu_alloc_maps import RescanGPUAllocMapsManifest
from ai.backend.manager.bgtask.types import ManagerBgtaskName
from ai.backend.manager.data.agent.types import AgentData
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.repositories.agent.query import QueryConditions, QueryOrders

from ..agent import (
    ADMIN_PERMISSIONS,
    AgentRow,
    AgentStatus,
    agents,
    get_permission_ctx,
)
from ..base import (
    FilterExprArg,
    Item,
    OrderExprArg,
    PaginatedConnectionField,
    PaginatedList,
    generate_sql_info_for_gql_connection,
    privileged_mutation,
    set_if_set,
    simple_db_mutate,
)
from ..gql_models.kernel import ComputeContainer
from ..gql_relay import AsyncNode, Connection, ConnectionResolverResult
from ..group import AssocGroupUserRow
from ..kernel import KernelRow
from ..keypair import keypairs
from ..minilang.ordering import OrderSpecItem, QueryOrderParser
from ..minilang.queryfilter import FieldSpecItem, QueryFilterParser
from ..rbac import (
    ScopeType,
)
from ..rbac.context import ClientContext
from ..rbac.permission_defs import AgentPermission
from ..user import UserRole, users
from .base import UUIDFloatMap
from .fields import AgentPermissionField
from .kernel import KernelConnection, KernelNode

if TYPE_CHECKING:
    from ..gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "Agent",
    "AgentNode",
    "AgentConnection",
    "AgentSummary",
    "AgentList",
    "AgentSummaryList",
    "ModifyAgent",
    "ModifyAgentInput",
)

_queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
    "id": ("id", None),
    "status": ("status", AgentStatus),
    "status_changed": ("status_changed", dtparse),
    "region": ("region", None),
    "scaling_group": ("scaling_group", None),
    "schedulable": ("schedulable", None),
    "addr": ("addr", None),
    "first_contact": ("first_contact", dtparse),
    "lost_at": ("lost_at", dtparse),
    "version": ("version", None),
}

_queryorder_colmap: Mapping[str, OrderSpecItem] = {
    "id": ("id", None),
    "status": ("status", None),
    "status_changed": ("status_changed", None),
    "region": ("region", None),
    "scaling_group": ("scaling_group", None),
    "schedulable": ("schedulable", None),
    "first_contact": ("first_contact", None),
    "lost_at": ("lost_at", None),
    "version": ("version", None),
    "available_slots": ("available_slots", None),
    "occupied_slots": ("occupied_slots", None),
}


async def _resolve_gpu_alloc_map(ctx: GraphQueryContext, agent_id: AgentId) -> dict[str, float]:
    raw_alloc_map = await ctx.valkey_stat.get_gpu_allocation_map(str(agent_id))
    if raw_alloc_map:
        return UUIDFloatMap.parse_value({k: float(v) for k, v in raw_alloc_map.items()})
    return {}


class AgentNode(graphene.ObjectType):
    class Meta:
        interfaces = (AsyncNode,)
        description = "Added in 24.12.0."

    row_id = graphene.String()
    status = graphene.String()
    status_changed = GQLDateTime()
    region = graphene.String()
    scaling_group = graphene.String()
    schedulable = graphene.Boolean()
    available_slots = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    addr = graphene.String(description="Agent's address with port. (bind/advertised host:port)")
    architecture = graphene.String()
    first_contact = GQLDateTime()
    lost_at = GQLDateTime()
    live_stat = graphene.JSONString()
    version = graphene.String()
    compute_plugins = graphene.JSONString()
    hardware_metadata = graphene.JSONString()
    auto_terminate_abusing_kernel = graphene.Boolean()
    local_config = graphene.JSONString()
    container_count = graphene.Int()
    gpu_alloc_map = UUIDFloatMap(description="Added in 25.4.0.")

    kernel_nodes = PaginatedConnectionField(
        KernelConnection,
    )

    permissions = graphene.List(
        AgentPermissionField,
        description=f"Added in 24.12.0. One of {[val.value for val in AgentPermission]}.",
    )

    @classmethod
    async def get_node(cls, info: graphene.ResolveInfo, id: str) -> Optional[Self]:
        graphene_ctx: GraphQueryContext = info.context
        _, raw_agent_id = AsyncNode.resolve_global_id(info, id)
        condition = [QueryConditions.by_ids([AgentId(raw_agent_id)])]
        agent_list = await graphene_ctx.agent_repository.list_data(condition)
        if len(agent_list) == 0:
            return None
        return cls.from_data(agent_list[0])

    @classmethod
    def from_data(cls, data: AgentData) -> Self:
        return cls(
            id=data.id,
            row_id=data.id,
            status=data.status.name,
            status_changed=data.status_changed,
            region=data.region,
            scaling_group=data.scaling_group,
            schedulable=data.schedulable,
            available_slots=data.available_slots.to_json(),
            occupied_slots=data.actual_occupied_slots.to_json(),
            addr=data.addr,
            architecture=data.architecture,
            first_contact=data.first_contact,
            lost_at=data.lost_at,
            version=data.version,
            compute_plugins=data.compute_plugins,
            auto_terminate_abusing_kernel=data.auto_terminate_abusing_kernel,
        )

    async def resolve_kernel_nodes(
        self, info: graphene.ResolveInfo
    ) -> ConnectionResolverResult[KernelNode]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, KernelNode.batch_load_by_agent_id)
        result = await loader.load(self.id)
        return ConnectionResolverResult(result, None, None, None, len(result))

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, self.batch_load_live_stat)
        return await loader.load(self.id)

    async def resolve_gpu_alloc_map(self, info: graphene.ResolveInfo) -> dict[str, float]:
        return await _resolve_gpu_alloc_map(info.context, self.id)

    async def resolve_hardware_metadata(
        self,
        info: graphene.ResolveInfo,
    ) -> Optional[Mapping[str, HardwareMetadata]]:
        if self.status != AgentStatus.ALIVE.name:
            return None
        graph_ctx: GraphQueryContext = info.context
        return await graph_ctx.registry.gather_agent_hwinfo(self.id)

    async def resolve_local_config(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        return {
            "agent": {
                "auto_terminate_abusing_kernel": self.auto_terminate_abusing_kernel,
            },
        }

    async def resolve_container_count(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, self.batch_load_container_count)
        return await loader.load(self.id)

    @classmethod
    async def batch_load_live_stat(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[Any]:
        return await ctx.valkey_stat.get_agent_statistics_batch(list(agent_ids))

    @classmethod
    async def batch_load_container_count(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[int]:
        return await ctx.valkey_stat.get_agent_container_counts_batch(list(agent_ids))

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        scope: ScopeType,
        permission: AgentPermission,
        filter_expr: Optional[str] = None,
        order_expr: Optional[str] = None,
        offset: Optional[int] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        before: Optional[str] = None,
        last: Optional[int] = None,
    ) -> ConnectionResolverResult:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(_queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(_queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            AgentRow,
            AgentRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            if user["role"] != UserRole.SUPERADMIN:
                client_ctx = ClientContext(
                    graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
                )
                permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope, permission)
                cond = permission_ctx.query_condition
                if cond is None:
                    return ConnectionResolverResult([], cursor, pagination_order, page_size, 0)
                permission_getter = permission_ctx.calculate_final_permission
                query = query.where(cond)
                cnt_query = cnt_query.where(cond)
            else:

                async def all_permissions(row):
                    return ADMIN_PERMISSIONS

                permission_getter = all_permissions
            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                agent_rows = (await db_session.scalars(query)).all()
                total_cnt = await db_session.scalar(cnt_query)
        agent_ids: list[AgentId] = []

        agent_permissions: dict[AgentId, list[AgentPermission]] = {}
        for row in agent_rows:
            agent_ids.append(row.id)
            permissions = await permission_getter(row)
            agent_permissions[row.id] = list(permissions)
        list_order = {agent_id: idx for idx, agent_id in enumerate(agent_ids)}
        condition = [QueryConditions.by_ids(agent_ids)]
        agent_list = await graph_ctx.agent_repository.list_data(condition)

        result: list[AgentNode] = []
        for agent in sorted(agent_list, key=lambda obj: list_order[obj.id]):
            agent_node = cls.from_data(agent)
            agent_node.permissions = agent_permissions.get(agent.id, [])
            result.append(agent_node)

        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)


class AgentConnection(Connection):
    class Meta:
        node = AgentNode
        description = "Added in 24.12.0."


### Legacy


class Agent(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    status = graphene.String()
    status_changed = GQLDateTime()
    region = graphene.String()
    scaling_group = graphene.String()
    schedulable = graphene.Boolean()
    available_slots = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    addr = graphene.String()  # bind/advertised host:port
    architecture = graphene.String()
    first_contact = GQLDateTime()
    lost_at = GQLDateTime()
    live_stat = graphene.JSONString()
    version = graphene.String()
    compute_plugins = graphene.JSONString()
    hardware_metadata = graphene.JSONString()
    auto_terminate_abusing_kernel = graphene.Boolean()
    local_config = graphene.JSONString()
    container_count = graphene.Int()
    gpu_alloc_map = UUIDFloatMap(description="Added in 25.4.0.")

    # Legacy fields
    mem_slots = graphene.Int()
    cpu_slots = graphene.Float()
    gpu_slots = graphene.Float()
    tpu_slots = graphene.Float()
    used_mem_slots = graphene.Int()
    used_cpu_slots = graphene.Float()
    used_gpu_slots = graphene.Float()
    used_tpu_slots = graphene.Float()
    cpu_cur_pct = graphene.Float()
    mem_cur_bytes = graphene.Float()

    compute_containers = graphene.List(ComputeContainer, status=graphene.String())

    @classmethod
    def from_data(cls, data: AgentData) -> Self:
        mega = 2**20
        return cls(
            id=data.id,
            status=data.status.name,
            status_changed=data.status_changed,
            region=data.region,
            scaling_group=data.scaling_group,
            schedulable=data.schedulable,
            available_slots=data.available_slots.to_json(),
            occupied_slots=data.actual_occupied_slots.to_json(),
            addr=data.addr,
            architecture=data.architecture,
            first_contact=data.first_contact,
            lost_at=data.lost_at,
            version=data.version,
            compute_plugins=data.compute_plugins,
            auto_terminate_abusing_kernel=False,  # legacy field
            # legacy fields
            mem_slots=data.available_slots.get("mem", 0) // mega,
            cpu_slots=data.available_slots.get("cpu", 0),
            gpu_slots=data.available_slots.get("cuda.device", 0),
            tpu_slots=data.available_slots.get("tpu.device", 0),
            used_mem_slots=data.actual_occupied_slots.get("mem", 0) // mega,
            used_cpu_slots=float(data.actual_occupied_slots.get("cpu", 0)),
            used_gpu_slots=float(data.actual_occupied_slots.get("cuda.device", 0)),
            used_tpu_slots=float(data.actual_occupied_slots.get("tpu.device", 0)),
        )

    async def resolve_compute_containers(
        self, info: graphene.ResolveInfo, *, status: Optional[str] = None
    ) -> list[ComputeContainer]:
        ctx: GraphQueryContext = info.context
        _status = KernelStatus[status] if status is not None else None
        loader = ctx.dataloader_manager.get_loader_by_func(
            ctx,
            ComputeContainer.batch_load_by_agent_id,
            status=_status,
        )
        return await loader.load(self.id)

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, Agent.batch_load_live_stat)
        return await loader.load(self.id)

    async def resolve_cpu_cur_pct(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, Agent.batch_load_cpu_cur_pct)
        return await loader.load(self.id)

    async def resolve_mem_cur_bytes(self, info: graphene.ResolveInfo) -> Any:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, Agent.batch_load_mem_cur_bytes)
        return await loader.load(self.id)

    async def resolve_hardware_metadata(
        self,
        info: graphene.ResolveInfo,
    ) -> Optional[Mapping[str, HardwareMetadata]]:
        if self.status != AgentStatus.ALIVE.name:
            return None
        graph_ctx: GraphQueryContext = info.context
        return await graph_ctx.registry.gather_agent_hwinfo(self.id)

    async def resolve_local_config(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        return {
            "agent": {
                "auto_terminate_abusing_kernel": self.auto_terminate_abusing_kernel,
            },
        }

    async def resolve_container_count(self, info: graphene.ResolveInfo) -> int:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, Agent.batch_load_container_count)
        return await loader.load(self.id)

    async def resolve_gpu_alloc_map(self, info: graphene.ResolveInfo) -> dict[str, float]:
        return await _resolve_gpu_alloc_map(info.context, self.id)

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "id": ("id", None),
        "status": ("status", AgentStatus),
        "status_changed": ("status_changed", dtparse),
        "region": ("region", None),
        "scaling_group": ("scaling_group", None),
        "schedulable": ("schedulable", None),
        "addr": ("addr", None),
        "first_contact": ("first_contact", dtparse),
        "lost_at": ("lost_at", dtparse),
        "version": ("version", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "id": ("id", None),
        "status": ("status", None),
        "status_changed": ("status_changed", None),
        "region": ("region", None),
        "scaling_group": ("scaling_group", None),
        "schedulable": ("schedulable", None),
        "first_contact": ("first_contact", None),
        "lost_at": ("lost_at", None),
        "version": ("version", None),
        "available_slots": ("available_slots", None),
        "occupied_slots": ("occupied_slots", None),
    }

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        scaling_group: Optional[str] = None,
        raw_status: Optional[str | AgentStatus] = None,
        filter: Optional[str] = None,
    ) -> int:
        if isinstance(raw_status, str):
            status_list = [AgentStatus[s] for s in raw_status.split(",")]
        elif isinstance(raw_status, AgentStatus):
            status_list = [raw_status]
        query = sa.select([sa.func.count()]).select_from(agents)
        if scaling_group is not None:
            query = query.where(agents.c.scaling_group == scaling_group)
        if raw_status is not None:
            query = query.where(agents.c.status.in_(status_list))
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        scaling_group: Optional[str] = None,
        raw_status: Optional[str] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[Agent]:
        if isinstance(raw_status, str):
            status_list = [AgentStatus[s] for s in raw_status.split(",")]
        elif isinstance(raw_status, AgentStatus):
            status_list = [raw_status]
        query = sa.select([agents]).select_from(agents).limit(limit).offset(offset)
        if scaling_group is not None:
            query = query.where(agents.c.scaling_group == scaling_group)
        if raw_status is not None:
            query = query.where(agents.c.status.in_(status_list))
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(
                agents.c.status.asc(),
                agents.c.scaling_group.asc(),
                agents.c.id.asc(),
            )
        agent_ids: list[AgentId] = []
        async with graph_ctx.db.begin_readonly() as conn:
            async for row in await conn.stream(query):
                agent_ids.append(row.id)
        list_order = {agent_id: idx for idx, agent_id in enumerate(agent_ids)}
        condition = [QueryConditions.by_ids(agent_ids)]
        agent_list = await graph_ctx.agent_repository.list_data(condition)
        return [
            cls.from_data(agent) for agent in sorted(agent_list, key=lambda obj: list_order[obj.id])
        ]

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        scaling_group: Optional[str] = None,
        raw_status: Optional[str] = None,
    ) -> Sequence[Agent]:
        conditions = []
        if scaling_group is not None:
            conditions.append(QueryConditions.by_scaling_group(scaling_group))
        if raw_status is not None:
            conditions.append(QueryConditions.by_statuses([AgentStatus[raw_status]]))

        agent_list = await graph_ctx.agent_repository.list_data(conditions)
        return [cls.from_data(agent) for agent in agent_list]

    @classmethod
    async def batch_load(
        cls,
        graph_ctx: GraphQueryContext,
        agent_ids: Sequence[AgentId],
        *,
        raw_status: Optional[str] = None,
    ) -> Sequence[Agent | None]:
        condition = [QueryConditions.by_ids(agent_ids)]
        order = [QueryOrders.id(ascending=True)]
        if raw_status is not None:
            condition.append(QueryConditions.by_statuses([AgentStatus[raw_status]]))
        agent_list = await graph_ctx.agent_repository.list_data(
            conditions=condition, order_by=order
        )
        return [cls.from_data(agent) for agent in agent_list]

    @classmethod
    async def batch_load_live_stat(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[Any]:
        return await ctx.valkey_stat.get_agent_statistics_batch(list(agent_ids))

    @classmethod
    async def batch_load_cpu_cur_pct(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[Any]:
        ret = []
        for stat in await cls.batch_load_live_stat(ctx, agent_ids):
            if stat is not None:
                try:
                    ret.append(float(stat["node"]["cpu_util"]["pct"]))
                except (KeyError, TypeError, ValueError):
                    ret.append(0.0)
            else:
                ret.append(0.0)
        return ret

    @classmethod
    async def batch_load_mem_cur_bytes(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[Any]:
        ret = []
        for stat in await cls.batch_load_live_stat(ctx, agent_ids):
            if stat is not None:
                try:
                    ret.append(float(stat["node"]["mem"]["current"]))
                except (KeyError, TypeError, ValueError):
                    ret.append(0)
            else:
                ret.append(0)
        return ret

    @classmethod
    async def batch_load_container_count(
        cls, ctx: GraphQueryContext, agent_ids: Sequence[str]
    ) -> Sequence[int]:
        return await ctx.valkey_stat.get_agent_container_counts_batch(list(agent_ids))


async def _query_domain_groups_by_ak(
    db_conn: SAConnection,
    access_key: str,
    domain_name: str | None,
) -> tuple[str, list[uuid.UUID]]:
    kp_user_join = sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
    if domain_name is None:
        domain_query = (
            sa.select([users.c.uuid, users.c.domain_name])
            .select_from(kp_user_join)
            .where(keypairs.c.access_key == access_key)
        )
        row = (await db_conn.execute(domain_query)).first()
        user_domain = row.domain_name
        user_id = row.uuid
        group_join = AssocGroupUserRow
        group_cond = AssocGroupUserRow.user_id == user_id
    else:
        user_domain = domain_name
        group_join = kp_user_join.join(
            AssocGroupUserRow,
            AssocGroupUserRow.user_id == users.c.uuid,
        )
        group_cond = keypairs.c.access_key == access_key
    query = sa.select(AssocGroupUserRow.group_id).select_from(group_join).where(group_cond)
    rows = (await db_conn.execute(query)).fetchall()
    group_ids = [row.group_id for row in rows]
    return user_domain, group_ids


async def _append_sgroup_from_clause(
    graph_ctx: GraphQueryContext,
    query: sa.sql.Select,
    access_key: str,
    domain_name: str | None,
    scaling_group: str | None = None,
) -> sa.sql.Select:
    from ..scaling_group import query_allowed_sgroups

    if scaling_group is not None:
        query = query.where(AgentRow.scaling_group == scaling_group)
    else:
        async with graph_ctx.db.begin_readonly() as conn:
            domain_name, group_ids = await _query_domain_groups_by_ak(conn, access_key, domain_name)
            sgroups = await query_allowed_sgroups(conn, domain_name, group_ids, access_key)
            names = [sgroup["name"] for sgroup in sgroups]
        query = query.where(AgentRow.scaling_group.in_(names))
    return query


class AgentList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(Agent, required=True)


class AgentSummary(graphene.ObjectType):
    """
    A schema for normal users.
    """

    class Meta:
        interfaces = (Item,)

    status = graphene.String()
    scaling_group = graphene.String()
    schedulable = graphene.Boolean()
    available_slots = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    architecture = graphene.String()

    @classmethod
    def from_data(cls, data: AgentData) -> Self:
        return cls(
            id=data.id,
            status=data.status.name,
            scaling_group=data.scaling_group,
            schedulable=data.schedulable,
            available_slots=data.available_slots.to_json(),
            occupied_slots=data.actual_occupied_slots.to_json(),
            architecture=data.architecture,
        )

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "id": ("id", None),
        "status": ("status", AgentStatus),
        "scaling_group": ("scaling_group", None),
        "schedulable": ("schedulable", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "id": ("id", None),
        "status": ("status", None),
        "scaling_group": ("scaling_group", None),
        "schedulable": ("schedulable", None),
        "available_slots": ("available_slots", None),
        "occupied_slots": ("occupied_slots", None),
    }

    @classmethod
    async def batch_load(
        cls,
        graph_ctx: GraphQueryContext,
        agent_ids: Sequence[AgentId],
        *,
        access_key: AccessKey,
        domain_name: Optional[str] = None,
        raw_status: Optional[str] = None,
        scaling_group: Optional[str] = None,
    ) -> Sequence[Optional[Self]]:
        query = (
            sa.select(AgentRow)
            .select_from(
                sa.join(
                    AgentRow,
                    KernelRow,
                    sa.and_(
                        AgentRow.id == KernelRow.agent,
                        KernelRow.status.in_(KernelStatus.resource_occupied_statuses()),
                    ),
                    isouter=True,
                )
            )
            .where(AgentRow.id.in_(agent_ids))
            .options(contains_eager(AgentRow.kernels))
            .order_by(
                AgentRow.id,
            )
        )
        if raw_status is not None:
            query = query.where(AgentRow.status == AgentStatus[raw_status])
        query = await _append_sgroup_from_clause(
            graph_ctx, query, access_key, domain_name, scaling_group
        )
        async with graph_ctx.db.begin_readonly_session() as session:
            result = await session.scalars(query)
            agent_list = result.unique().all()
            return [cls.from_data(agent.to_data()) for agent in agent_list]

    @classmethod
    async def load_count(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        access_key: str,
        domain_name: str | None = None,
        scaling_group: str | None = None,
        raw_status: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> int:
        query = sa.select(sa.func.count()).select_from(AgentRow)
        query = await _append_sgroup_from_clause(
            graph_ctx, query, access_key, domain_name, scaling_group
        )

        if raw_status is not None:
            query = query.where(AgentRow.status == AgentStatus[raw_status])
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with graph_ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        access_key: str,
        domain_name: str | None = None,
        scaling_group: str | None = None,
        raw_status: Optional[str] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[Self]:
        query = sa.select(AgentRow)

        if raw_status is not None:
            query = query.where(AgentRow.status == AgentStatus[raw_status])
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(
                AgentRow.status.asc(),
                AgentRow.scaling_group.asc(),
                AgentRow.id.asc(),
            )
        query = (
            query.select_from(
                sa.join(
                    AgentRow,
                    KernelRow,
                    sa.and_(
                        AgentRow.id == KernelRow.agent,
                        KernelRow.status.in_(KernelStatus.resource_occupied_statuses()),
                    ),
                    isouter=True,
                )
            )
            .options(contains_eager(AgentRow.kernels))
            .limit(limit)
            .offset(offset)
        )
        query = await _append_sgroup_from_clause(
            graph_ctx, query, access_key, domain_name, scaling_group
        )
        agent_ids: list[AgentId] = []
        async with graph_ctx.db.begin_readonly_session() as db_session:
            result = await db_session.scalars(query)
            rows = result.unique().all()
            for row in rows:
                agent_ids.append(row.id)

        list_order = {agent_id: idx for idx, agent_id in enumerate(agent_ids)}
        condition = [QueryConditions.by_ids(agent_ids)]
        agent_list = await graph_ctx.agent_repository.list_data(condition)
        return [
            cls.from_data(agent) for agent in sorted(agent_list, key=lambda obj: list_order[obj.id])
        ]


class AgentSummaryList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(AgentSummary, required=True)


class ModifyAgentInput(graphene.InputObjectType):
    schedulable = graphene.Boolean(required=False, default=True)
    scaling_group = graphene.String(required=False)


class ModifyAgent(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        id = graphene.String(required=True)
        props = ModifyAgentInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    @privileged_mutation(
        UserRole.SUPERADMIN,
        lambda id, **kwargs: (None, id),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        id: str,
        props: ModifyAgentInput,
    ) -> ModifyAgent:
        graph_ctx: GraphQueryContext = info.context
        data: dict[str, Any] = {}
        set_if_set(props, data, "schedulable")
        set_if_set(props, data, "scaling_group")
        # TODO: Need to skip the following RPC call if the agent is not alive, or timeout.
        if (scaling_group := data.get("scaling_group")) is not None:
            await graph_ctx.registry.update_scaling_group(id, scaling_group)

        update_query = sa.update(agents).values(data).where(agents.c.id == id)
        return await simple_db_mutate(cls, graph_ctx, update_query)


class RescanGPUAllocMaps(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)

    class Meta:
        description = "Added in 25.4.0."

    class Arguments:
        agent_id = graphene.String(
            description="Agent ID to rescan GPU alloc map",
            required=True,
        )

    task_id = graphene.UUID()

    @classmethod
    @privileged_mutation(
        UserRole.SUPERADMIN,
        lambda id, **kwargs: (None, id),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        agent_id: str,
    ) -> RescanGPUAllocMaps:
        log.info("rescanning GPU alloc maps for agent {}", agent_id)
        graph_ctx: GraphQueryContext = info.context

        manifest = RescanGPUAllocMapsManifest(agent_id=AgentId(agent_id))
        task_id = await graph_ctx.background_task_manager.start_retriable(
            ManagerBgtaskName.RESCAN_GPU_ALLOC_MAPS,
            manifest,
        )
        return RescanGPUAllocMaps(task_id=task_id)



from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.errors.common import ObjectNotFound


class DefinitionFileNotFound(ObjectNotFound):
    object_name = "definition-file"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class EndpointNotFound(ObjectNotFound):
    object_name = "endpoint"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class UserNotFoundInDeployment(ObjectNotFound):
    object_name = "user in deployment"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class DeploymentHasNoTargetRevision(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/deployment-has-no-target-revision"
    error_title = "Deployment has no target revision."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



import asyncio
import logging
from typing import Awaitable, Callable, Self

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.loader.loader_chain import LoaderChain
from ai.backend.manager.config.watchers.etcd import EtcdConfigWatcher

from .loader.legacy_etcd_loader import LegacyEtcdLoader
from .unified import ManagerUnifiedConfig

SharedConfigChangeCallback = Callable[[ManagerUnifiedConfig], Awaitable[None]]

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ManagerConfigProvider:
    _loader: LoaderChain
    _config: ManagerUnifiedConfig
    _etcd_watcher: EtcdConfigWatcher
    _etcd_watcher_task: asyncio.Task[None]
    # TODO: Remove `_legacy_etcd_config_loader` when legacy etcd methods are removed
    _legacy_etcd_config_loader: LegacyEtcdLoader

    def __init__(
        self,
        loader: LoaderChain,
        config: ManagerUnifiedConfig,
        etcd_watcher: EtcdConfigWatcher,
        legacy_etcd_config_loader: LegacyEtcdLoader,
    ) -> None:
        self._loader = loader
        self._config = config
        self._etcd_watcher = etcd_watcher
        self._legacy_etcd_config_loader = legacy_etcd_config_loader
        self._etcd_watcher_task = asyncio.create_task(self._run_watcher())

    @classmethod
    async def create(
        cls,
        loader: LoaderChain,
        etcd_watcher: EtcdConfigWatcher,
        legacy_etcd_config_loader: LegacyEtcdLoader,
    ) -> Self:
        raw_config = await loader.load()
        config = ManagerUnifiedConfig.model_validate(raw_config, by_name=True)
        return cls(loader, config, etcd_watcher, legacy_etcd_config_loader)

    @property
    def config(self) -> ManagerUnifiedConfig:
        return self._config

    def reload(self, config: ManagerUnifiedConfig) -> None:
        self._config = config

    @property
    def legacy_etcd_config_loader(self) -> LegacyEtcdLoader:
        return self._legacy_etcd_config_loader

    async def _run_watcher(self) -> None:
        async for event in self._etcd_watcher.watch():
            raw_config = await self._loader.load()
            self._config = ManagerUnifiedConfig.model_validate(raw_config, by_name=True)
            log.debug("config reloaded due to etcd event.")

    async def terminate(self) -> None:
        if self._etcd_watcher_task:
            self._etcd_watcher_task.cancel()
            try:
                await self._etcd_watcher_task
            except asyncio.CancelledError:
                pass



"""
Service endpoint and model service-related exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)

from .common import ObjectNotFound


class NoUpdatesToApply(BackendAIError):
    """Raised when there are no updates to apply to an endpoint."""

    def __init__(self, message: str = "No updates to apply") -> None:
        super().__init__(message)

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class EndpointNotFound(ObjectNotFound):
    object_name = "endpoint"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ModelDefinitionNotFound(ObjectNotFound):
    object_name = "model_definition"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ScalingImpossible(BackendAIError, web.HTTPBadRequest):
    error_title = (
        "Scaling operation cannot be performed due to insufficient resources or constraints."
    )

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.CONFLICT,
        )


class AutoScalingRuleNotFound(ObjectNotFound):
    object_name = "auto_scaling_rule"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class RoutingNotFound(ObjectNotFound):
    object_name = "routing"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ROUTE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class EndpointTokenNotFound(ObjectNotFound):
    object_name = "endpoint_token"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ENDPOINT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ModelServiceNotFound(ObjectNotFound):
    object_name = "model service"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class RouteNotFound(ObjectNotFound):
    object_name = "route"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ROUTE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ModelServiceDependencyNotCleared(BackendAIError, web.HTTPBadRequest):
    error_title = "Cannot delete model VFolders bound to alive model services."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.CONFLICT,
        )


class AppServiceStartFailed(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/app-service-start-failed"
    error_title = "Failed to start the application service."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.START,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )



import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from ai.backend.manager.api.gql.base import IntFilter, StringFilter
from ai.backend.manager.data.artifact.types import (
    ArtifactRemoteStatus,
    ArtifactRevisionOrderField,
    ArtifactStatus,
)


@dataclass
class ArtifactRevisionOrderingOptions:
    """Ordering options for artifact revision queries."""

    order_by: list[tuple[ArtifactRevisionOrderField, bool]] = field(
        default_factory=lambda: [(ArtifactRevisionOrderField.CREATED_AT, False)]
    )  # (field, desc)


class ArtifactStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class ArtifactStatusFilter:
    """Status filter with operation type and values."""

    type: ArtifactStatusFilterType
    values: list[ArtifactStatus]


class ArtifactRemoteStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class ArtifactRemoteStatusFilter:
    """Remote status filter with operation type and values."""

    type: ArtifactRemoteStatusFilterType
    values: list[ArtifactRemoteStatus]


@dataclass
class ArtifactRevisionFilterOptions:
    """Filtering options for artifact revisions."""

    artifact_id: Optional[uuid.UUID] = None
    status_filter: Optional[ArtifactStatusFilter] = None
    remote_status_filter: Optional[ArtifactRemoteStatusFilter] = None
    version_filter: Optional[StringFilter] = None
    size_filter: Optional[IntFilter] = None

    # Logical operations
    AND: Optional[list["ArtifactRevisionFilterOptions"]] = None
    OR: Optional[list["ArtifactRevisionFilterOptions"]] = None
    NOT: Optional[list["ArtifactRevisionFilterOptions"]] = None



from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.types import (
    ModelRevisionData,
)
from ai.backend.manager.services.deployment.actions.model_revision.base import (
    ModelRevisionBaseAction,
)


@dataclass
class GetRevisionByIdAction(ModelRevisionBaseAction):
    revision_id: UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.revision_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get"


@dataclass
class GetRevisionByIdActionResult(BaseActionResult):
    data: ModelRevisionData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id)



from __future__ import annotations

from textwrap import dedent


def dedent_strip(text: str) -> str:
    """
    Apply textwrap.dedent and strip to remove both indentation and leading/trailing whitespace.
    This is useful for GraphQL descriptions to ensure clean output in schema introspection.
    """
    return dedent(text).strip()



from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class ObjectStorageNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/object-storage-not-found"
    error_title = "Object Storage Not Found"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.OBJECT_STORAGE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ObjectStorageOperationNotSupported(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/object-storage-operation-not-supported"
    error_title = "Object Storage Operation Not Supported"

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.OBJECT_STORAGE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



from __future__ import annotations

import logging

from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.data.image.types import InstalledImageInfo
from ai.backend.common.types import AgentId
from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AgentStatefulSource:
    """
    Stateful source for agent-related operations.
    Uses simple JSON serialization with short TTL for caching.
    """

    _valkey_image: ValkeyImageClient

    def __init__(
        self,
        valkey_image: ValkeyImageClient,
    ) -> None:
        self._valkey_image = valkey_image

    async def read_agent_installed_images(self, agent_id: AgentId) -> list[InstalledImageInfo]:
        """Read installed images for the given agent IDs."""
        return await self._valkey_image.get_agent_installed_images(agent_id)



"""create global roles

Revision ID: 09206ac04fd3
Revises: de1032a11cca
Create Date: 2025-09-22 19:44:13.067238

"""

import enum
import uuid
from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection, Row

from ai.backend.manager.models.base import EnumValueType, IDColumn
from ai.backend.manager.models.rbac_models.migration.enums import (
    EntityType,
    OperationType,
)
from ai.backend.manager.models.rbac_models.migration.models import (
    get_user_roles_table,
    mapper_registry,
)
from ai.backend.manager.models.rbac_models.migration.types import (
    UserRoleCreateInput,
)
from ai.backend.manager.models.rbac_models.migration.user import (
    get_monitor_role_creation_input,
    get_superadmin_role_creation_input,
)
from ai.backend.manager.models.rbac_models.migration.utils import (
    PermissionUpdateUtil,
    insert_skip_on_conflict,
)

# revision identifiers, used by Alembic.
revision = "09206ac04fd3"
down_revision = "de1032a11cca"
branch_labels = None
depends_on = None


class UserRole(enum.StrEnum):
    """
    User's role.
    """

    SUPERADMIN = "superadmin"
    ADMIN = "admin"
    USER = "user"
    MONITOR = "monitor"


ENTITY_TYPES = {
    EntityType.USER,
    EntityType.PROJECT,
    EntityType.DOMAIN,
    EntityType.VFOLDER,
}


class Tables:
    @staticmethod
    def get_users_table() -> sa.Table:
        users_table = sa.Table(
            "users",
            mapper_registry.metadata,
            IDColumn("uuid"),
            sa.Column("username", sa.String(length=64), unique=True),
            sa.Column("role", EnumValueType(UserRole), default=UserRole.USER),
            extend_existing=True,
        )
        return users_table


class RoleCreator:
    @classmethod
    def _create_superadmin_role(cls, db_conn: Connection) -> uuid.UUID:
        role_input = get_superadmin_role_creation_input()
        role_id = PermissionUpdateUtil.get_or_create_role(db_conn, role_input)
        permission_group_id, exist = PermissionUpdateUtil.get_or_create_global_permission_group(
            db_conn, role_id
        )
        if not exist:
            PermissionUpdateUtil.create_permissions(
                db_conn, permission_group_id, ENTITY_TYPES, OperationType.admin_operations()
            )
        return role_id

    @classmethod
    def _create_monitor_role(cls, db_conn: Connection) -> uuid.UUID:
        role_input = get_monitor_role_creation_input()
        role_id = PermissionUpdateUtil.get_or_create_role(db_conn, role_input)
        permission_group_id, exist = PermissionUpdateUtil.get_or_create_global_permission_group(
            db_conn, role_id
        )
        if not exist:
            PermissionUpdateUtil.create_permissions(
                db_conn, permission_group_id, ENTITY_TYPES, OperationType.monitor_operations()
            )
        return role_id

    @classmethod
    def _query_user_row_by_role(
        cls, db_conn: Connection, role: UserRole, offset: int, page_size: int
    ) -> list[Row]:
        """
        Query all user rows with pagination.
        """
        users_table = Tables.get_users_table()
        user_query = (
            sa.select(
                users_table.c.uuid,
                users_table.c.username,
                users_table.c.role,
            )
            .offset(offset)
            .limit(page_size)
            .where(users_table.c.role == role)
            .order_by(users_table.c.uuid)
        )
        return db_conn.execute(user_query).all()

    @classmethod
    def _map_role_to_users(
        cls, db_conn: Connection, user_role: UserRole, role_id: uuid.UUID
    ) -> None:
        user_roles_table = get_user_roles_table()
        offset = 0
        page_size = 1000
        while True:
            rows = cls._query_user_row_by_role(db_conn, user_role, offset, page_size)
            offset += page_size
            if not rows:
                break

            user_role_inputs: list[dict[str, Any]] = [
                UserRoleCreateInput(user_id=row.uuid, role_id=role_id).to_dict() for row in rows
            ]
            insert_skip_on_conflict(db_conn, user_roles_table, user_role_inputs)

    @classmethod
    def create_and_map_superadmin_role(cls, db_conn: Connection) -> None:
        superadmin_role_id = cls._create_superadmin_role(db_conn)
        cls._map_role_to_users(db_conn, UserRole.SUPERADMIN, superadmin_role_id)

    @classmethod
    def create_and_map_monitor_role(cls, db_conn: Connection) -> None:
        monitor_role_id = cls._create_monitor_role(db_conn)
        cls._map_role_to_users(db_conn, UserRole.MONITOR, monitor_role_id)


def upgrade() -> None:
    conn = op.get_bind()
    RoleCreator.create_and_map_superadmin_role(conn)
    RoleCreator.create_and_map_monitor_role(conn)


def downgrade() -> None:
    pass



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class RenameSessionAction(SessionAction):
    session_name: str
    new_name: str
    owner_access_key: AccessKey

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "rename"


@dataclass
class RenameSessionActionResult(BaseActionResult):
    session_data: SessionData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_data.id)



import logging
import secrets
import uuid
from http import HTTPStatus

import aiohttp
import tomli
from pydantic import HttpUrl
from yarl import URL

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager, ProgressReporter
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.defs.session import SESSION_PRIORITY_DEFAULT
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
)
from ai.backend.common.events.event_types.kernel.types import (
    KernelLifecycleEventReason,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SchedulingBroadcastEvent,
)
from ai.backend.common.events.hub import EventHub
from ai.backend.common.events.hub.propagators.bypass import AsyncBypassPropagator
from ai.backend.common.events.types import EventDomain
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import (
    AccessKey,
    ImageAlias,
    KernelEnqueueingConfig,
    SessionTypes,
    VFolderID,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.types import ModelServiceDefinition
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.model_serving.creator import EndpointCreator
from ai.backend.manager.data.model_serving.types import (
    CompactServiceInfo,
    ErrorInfo,
    RequesterCtx,
    RouteInfo,
    ServiceInfo,
)
from ai.backend.manager.data.model_serving.types import (
    EndpointTokenData as ServiceEndpointTokenData,
)
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.errors.service import (
    EndpointNotFound,
    ModelServiceNotFound,
    RouteNotFound,
)
from ai.backend.manager.errors.storage import UnexpectedStorageProxyResponseError
from ai.backend.manager.models.endpoint import (
    EndpointLifecycle,
    EndpointTokenRow,
)
from ai.backend.manager.models.routing import RouteStatus
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.vfolder import VFolderOwnershipType, VFolderRow
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.model_serving.admin_repository import (
    AdminModelServingRepository,
)
from ai.backend.manager.repositories.model_serving.repository import ModelServingRepository
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationSpec,
)
from ai.backend.manager.services.model_serving.actions.clear_error import (
    ClearErrorAction,
    ClearErrorActionResult,
)
from ai.backend.manager.services.model_serving.actions.create_model_service import (
    CreateModelServiceAction,
    CreateModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_model_service import (
    DeleteModelServiceAction,
    DeleteModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_route import (
    DeleteRouteAction,
    DeleteRouteActionResult,
)
from ai.backend.manager.services.model_serving.actions.dry_run_model_service import (
    DryRunModelServiceAction,
    DryRunModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.force_sync import (
    ForceSyncAction,
    ForceSyncActionResult,
)
from ai.backend.manager.services.model_serving.actions.generate_token import (
    GenerateTokenAction,
    GenerateTokenActionResult,
)
from ai.backend.manager.services.model_serving.actions.get_model_service_info import (
    GetModelServiceInfoAction,
    GetModelServiceInfoActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_errors import (
    ListErrorsAction,
    ListErrorsActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_model_service import (
    ListModelServiceAction,
    ListModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_endpoint import (
    ModifyEndpointAction,
    ModifyEndpointActionResult,
)
from ai.backend.manager.services.model_serving.actions.update_route import (
    UpdateRouteAction,
    UpdateRouteActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    GenericForbidden,
    InvalidAPIParameters,
)
from ai.backend.manager.sokovan.deployment.deployment_controller import DeploymentController
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.types import UserScope

log = BraceStyleAdapter(logging.getLogger(__name__))


class ModelServingService:
    _agent_registry: AgentRegistry
    _background_task_manager: BackgroundTaskManager
    _event_dispatcher: EventDispatcher
    _event_hub: EventHub
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider
    _repository: ModelServingRepository
    _admin_repository: AdminModelServingRepository

    _valkey_live: ValkeyLiveClient
    _deployment_controller: DeploymentController
    _scheduling_controller: SchedulingController

    def __init__(
        self,
        agent_registry: AgentRegistry,
        background_task_manager: BackgroundTaskManager,
        event_dispatcher: EventDispatcher,
        event_hub: EventHub,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        valkey_live: ValkeyLiveClient,
        repository: ModelServingRepository,
        admin_repository: AdminModelServingRepository,
        deployment_controller: DeploymentController,
        scheduling_controller: SchedulingController,
    ) -> None:
        self._agent_registry = agent_registry
        self._background_task_manager = background_task_manager
        self._event_dispatcher = event_dispatcher
        self._event_hub = event_hub
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._valkey_live = valkey_live
        self._repository = repository
        self._admin_repository = admin_repository
        self._deployment_controller = deployment_controller
        self._scheduling_controller = scheduling_controller
        # Map SessionStatus to legacy event names for backward compatibility
        self._status_to_event_name: dict[SessionStatus, str] = {
            SessionStatus.PENDING: "session_enqueued",
            SessionStatus.SCHEDULED: "session_scheduled",
            SessionStatus.PREPARING: "session_preparing",
            SessionStatus.PULLING: "session_preparing",
            SessionStatus.PREPARED: "session_prepared",
            SessionStatus.CREATING: "session_creating",
            SessionStatus.RUNNING: "session_started",
            SessionStatus.TERMINATING: "session_terminating",
            SessionStatus.TERMINATED: "session_terminated",
            SessionStatus.CANCELLED: "session_cancelled",
            SessionStatus.ERROR: "session_cancelled",
        }

    async def _fetch_file_from_storage_proxy(
        self,
        filename: str,
        model_vfolder_row: VFolderRow,
    ) -> bytes:
        vfid = model_vfolder_row.vfid
        folder_host = model_vfolder_row.host

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(folder_host)

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        return await manager_client.fetch_file_content(
            volume_name,
            str(vfid),
            f"./{filename}",
        )

    async def create(self, action: CreateModelServiceAction) -> CreateModelServiceActionResult:
        service_prepare_ctx = action.creator.model_service_prepare_ctx

        # Get model vfolder
        model_vfolder_row = await self._repository.get_vfolder_by_id(service_prepare_ctx.model_id)
        if not model_vfolder_row:
            raise InvalidAPIParameters("Model vfolder not found")
        if model_vfolder_row.ownership_type == VFolderOwnershipType.GROUP:
            raise InvalidAPIParameters(
                "Cannot create model service with the project type's vfolder"
            )

        try:
            chunks = await self._fetch_file_from_storage_proxy(
                "service-definition.toml", model_vfolder_row
            )
        except UnexpectedStorageProxyResponseError:
            chunks = None

        if chunks:
            raw_service_definition = chunks.decode("utf-8")
            service_definition = tomli.loads(raw_service_definition)

            definition = action.creator.runtime_variant
            if definition in service_definition:
                variant_def = ModelServiceDefinition.model_validate(service_definition[definition])
                if variant_def.resource_slots:
                    action.creator.config.resources = variant_def.resource_slots
                if variant_def.environment:
                    action.creator.image = variant_def.environment.image
                    action.creator.architecture = variant_def.environment.architecture
                if variant_def.environ:
                    if action.creator.config.environ:
                        action.creator.config.environ.update(variant_def.environ)
                    else:
                        action.creator.config.environ = variant_def.environ

        creation_config = action.creator.config.to_dict()
        creation_config["mounts"] = [
            service_prepare_ctx.model_id,
            *[m.vfid.folder_id for m in service_prepare_ctx.extra_mounts],
        ]
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.creator.config.model_mount_destination,
            **{
                m.vfid.folder_id: m.kernel_path.as_posix() for m in service_prepare_ctx.extra_mounts
            },
        }
        creation_config["mount_options"] = {
            m.vfid.folder_id: {"permission": m.mount_perm} for m in service_prepare_ctx.extra_mounts
        }
        sudo_session_enabled = action.creator.sudo_session_enabled

        # Resolve image row - EndpointRow constructor needs the actual ImageRow object
        if action.creator.image is None or action.creator.image.strip() == "":
            raise InvalidAPIParameters("Image must be specified for model service creation")
        if action.creator.architecture is None or action.creator.architecture.strip() == "":
            raise InvalidAPIParameters("Architecture must be specified for model service creation")
        image_row = await self._repository.resolve_image_for_endpoint_creation([
            ImageIdentifier(action.creator.image, action.creator.architecture),
            ImageAlias(action.creator.image),
        ])

        if action.creator.config.resources is None:
            raise InvalidAPIParameters("Resources must be specified for model service creation")

        # check if session is valid to be created
        await self._agent_registry.create_session(
            "",
            image_row.image_ref,
            UserScope(
                domain_name=action.creator.domain_name,
                group_id=service_prepare_ctx.group_id,
                user_uuid=service_prepare_ctx.owner_uuid,
                user_role=service_prepare_ctx.owner_role,
            ),
            service_prepare_ctx.owner_access_key,
            service_prepare_ctx.resource_policy,
            SessionTypes.INFERENCE,
            creation_config,
            action.creator.cluster_mode,
            action.creator.cluster_size,
            dry_run=True,  # Setting this to True will prevent actual session from being enqueued
            bootstrap_script=action.creator.bootstrap_script,
            startup_command=action.creator.startup_command,
            tag=action.creator.tag,
            callback_url=URL(action.creator.callback_url.unicode_string())
            if action.creator.callback_url
            else None,
            sudo_session_enabled=sudo_session_enabled,
        )

        # Check endpoint name uniqueness
        is_name_available = await self._repository.check_endpoint_name_uniqueness(
            action.creator.service_name
        )
        if not is_name_available:
            raise InvalidAPIParameters("Cannot create multiple services with same name")

        project_id = await self._repository.resolve_group_id(
            action.creator.domain_name, action.creator.group_name
        )
        if project_id is None:
            raise InvalidAPIParameters(f"Invalid group name {action.creator.group_name}")

        endpoint = EndpointCreator(
            action.creator.service_name,
            model_definition_path=service_prepare_ctx.model_definition_path,
            created_user=action.request_user_id,
            session_owner=service_prepare_ctx.owner_uuid,
            replicas=action.creator.replicas,
            image=image_row.id,
            model=service_prepare_ctx.model_id,
            domain=action.creator.domain_name,
            project=project_id,
            resource_group=service_prepare_ctx.scaling_group,
            resource_slots=action.creator.config.resources,
            cluster_mode=action.creator.cluster_mode,
            cluster_size=action.creator.cluster_size,
            extra_mounts=list(service_prepare_ctx.extra_mounts),
            model_mount_destination=action.creator.config.model_mount_destination,
            tag=action.creator.tag,
            startup_command=action.creator.startup_command,
            callback_url=URL(action.creator.callback_url.unicode_string())
            if action.creator.callback_url
            else None,
            environ=action.creator.config.environ,
            bootstrap_script=action.creator.bootstrap_script,
            resource_opts=action.creator.config.resource_opts,
            open_to_public=action.creator.open_to_public,
            runtime_variant=action.creator.runtime_variant,
        )

        endpoint_data = await self._repository.create_endpoint_validated(
            endpoint, self._agent_registry
        )
        endpoint_id = endpoint_data.id

        return CreateModelServiceActionResult(
            ServiceInfo(
                endpoint_id=endpoint_id,
                model_id=endpoint.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
                name=action.creator.service_name,
                model_definition_path=service_prepare_ctx.model_definition_path,
                replicas=endpoint.replicas,
                desired_session_count=endpoint.replicas,
                active_routes=[],
                service_endpoint=None,
                is_public=action.creator.open_to_public,
                runtime_variant=action.creator.runtime_variant,
            )
        )

    async def list_serve(self, action: ListModelServiceAction) -> ListModelServiceActionResult:
        endpoints = await self._repository.list_endpoints_by_owner_validated(
            action.session_owener_id, name=action.name
        )

        return ListModelServiceActionResult(
            data=[
                CompactServiceInfo(
                    id=endpoint.id,
                    name=endpoint.name,
                    replicas=endpoint.replicas,
                    desired_session_count=endpoint.replicas,
                    active_route_count=len([
                        r for r in endpoint.routings if r.status == RouteStatus.HEALTHY
                    ])
                    if endpoint.routings
                    else 0,
                    service_endpoint=HttpUrl(endpoint.url) if endpoint.url else None,
                    is_public=endpoint.open_to_public,
                )
                for endpoint in endpoints
            ]
        )

    async def check_requester_access(self, requester_ctx: RequesterCtx) -> None:
        if requester_ctx.is_authorized is False:
            raise GenericForbidden("Only authorized requests may have access key scopes.")

    async def delete(self, action: DeleteModelServiceAction) -> DeleteModelServiceActionResult:
        service_id = action.service_id

        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        # Determine lifecycle stage based on routes
        has_routes = endpoint_data.routings and len(endpoint_data.routings) > 0
        if has_routes:
            lifecycle_stage = EndpointLifecycle.DESTROYING
            replicas = 0
        else:
            lifecycle_stage = EndpointLifecycle.DESTROYED
            replicas = None

        # Update endpoint lifecycle
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            await self._admin_repository.update_endpoint_lifecycle_force(
                service_id, lifecycle_stage, replicas
            )
        else:
            await self._repository.update_endpoint_lifecycle_validated(
                service_id,
                lifecycle_stage,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
                replicas,
            )

        return DeleteModelServiceActionResult(success=True)

    async def dry_run(self, action: DryRunModelServiceAction) -> DryRunModelServiceActionResult:
        # TODO: Seperate background task definition and trigger into different layer
        service_prepare_ctx = action.model_service_prepare_ctx
        # Get user with keypair
        created_user = await self._repository.get_user_with_keypair(action.request_user_id)
        if not created_user:
            raise InvalidAPIParameters("User not found")

        image_row = await self._repository.resolve_image_for_endpoint_creation([
            ImageIdentifier(action.image, action.architecture),
            ImageAlias(action.image),
        ])

        creation_config = action.config.to_dict()
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.config.model_mount_destination
        }
        sudo_session_enabled = action.sudo_session_enabled

        # Build SessionCreationSpec for dry-run
        session_creation_id = secrets.token_urlsafe(16)
        session_name = f"model-eval-{session_creation_id}"

        # Prepare mounts and environ
        mounts = [
            service_prepare_ctx.model_id,
            *[m.vfid for m in service_prepare_ctx.extra_mounts],
        ]
        mount_map: dict[uuid.UUID | VFolderID, str] = {
            service_prepare_ctx.model_id: action.config.model_mount_destination,
        }
        for m in service_prepare_ctx.extra_mounts:
            mount_map[m.vfid] = str(m.kernel_path)
        mount_options = {
            m.vfid: {"permission": m.mount_perm} for m in service_prepare_ctx.extra_mounts
        }
        environ = creation_config.get("environ", {})

        # Create single kernel spec - cluster replication is handled by ClusterPreparerRule
        kernel_spec = KernelEnqueueingConfig(
            image_ref=image_row.image_ref,
            cluster_role="main",
            cluster_idx=1,
            local_rank=0,
            cluster_hostname="main1",
            creation_config={
                "mounts": mounts,
                "mount_map": mount_map,
                "mount_options": mount_options,
                "environ": environ,
                "resources": creation_config.get("resources", {}),
                "resource_opts": creation_config.get("resource_opts", {}),
            },
            bootstrap_script=action.bootstrap_script or "",
            startup_command=action.startup_command,
            # uid/main_gid/supplementary_gids are filled from context.container_user_info
            uid=None,
            main_gid=None,
            supplementary_gids=[],
        )

        session_spec = SessionCreationSpec(
            session_creation_id=session_creation_id,
            session_name=session_name,
            access_key=AccessKey(service_prepare_ctx.owner_access_key),
            user_scope=UserScope(
                domain_name=action.domain_name,
                group_id=service_prepare_ctx.group_id,
                user_uuid=created_user.uuid,
                user_role=created_user.role,
            ),
            session_type=SessionTypes.INFERENCE,
            cluster_mode=action.cluster_mode,
            cluster_size=action.cluster_size,
            priority=SESSION_PRIORITY_DEFAULT,
            resource_policy=service_prepare_ctx.resource_policy,
            kernel_specs=[kernel_spec],
            creation_spec={
                "mounts": mounts,
                "mount_map": mount_map,
                "mount_options": mount_options,
                "model_definition_path": service_prepare_ctx.model_definition_path,
                "runtime_variant": action.runtime_variant,
                "environ": environ,
                "scaling_group": service_prepare_ctx.scaling_group,
                "resources": creation_config.get("resources", {}),
                "resource_opts": creation_config.get("resource_opts", {}),
                "preopen_ports": None,
                "agent_list": None,
            },
            scaling_group=service_prepare_ctx.scaling_group,
            session_tag=action.tag,
            callback_url=URL(action.callback_url.unicode_string()) if action.callback_url else None,
            sudo_session_enabled=sudo_session_enabled,
        )

        # Enqueue session before starting background task to avoid race conditions
        session_id = await self._scheduling_controller.enqueue_session(session_spec)
        session_id_str = str(session_id)

        async def _task(reporter: ProgressReporter) -> None:
            # Use AsyncBypassPropagator to receive SchedulingBroadcastEvent
            propagator = AsyncBypassPropagator()
            try:
                self._event_hub.register_event_propagator(
                    propagator, aliases=[(EventDomain.SESSION, session_id_str)]
                )

                async for event in propagator.receive():
                    if isinstance(event, SchedulingBroadcastEvent):
                        status = SessionStatus(event.status_transition)
                        # Convert status to legacy event name for backward compatibility
                        event_name = self._status_to_event_name.get(status, event.status_transition)
                        task_message = {
                            "event": event_name,
                            "session_id": str(event.session_id),
                        }
                        await reporter.update(message=dump_json_str(task_message))

                        # When session becomes RUNNING, mark for termination and exit
                        if status == SessionStatus.RUNNING:
                            await self._scheduling_controller.mark_sessions_for_termination(
                                [session_id],
                                reason="DRY_RUN_COMPLETE",
                            )

                        # Exit loop on terminal states
                        if status.is_terminal():
                            break
            finally:
                self._event_hub.unregister_event_propagator(propagator.id())

        task_id = await self._background_task_manager.start(_task)
        return DryRunModelServiceActionResult(task_id)

    async def get_model_service_info(
        self, action: GetModelServiceInfoAction
    ) -> GetModelServiceInfoActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        return GetModelServiceInfoActionResult(
            ServiceInfo(
                endpoint_id=endpoint_data.id,
                model_id=endpoint_data.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint_data.extra_mounts],
                name=endpoint_data.name,
                model_definition_path=endpoint_data.model_definition_path,
                replicas=endpoint_data.replicas,
                desired_session_count=endpoint_data.replicas,
                active_routes=[
                    RouteInfo(
                        route_id=r.id,
                        session_id=r.session,
                        traffic_ratio=r.traffic_ratio,
                    )
                    for r in endpoint_data.routings
                ]
                if endpoint_data.routings
                else [],
                service_endpoint=HttpUrl(endpoint_data.url) if endpoint_data.url else None,
                is_public=endpoint_data.open_to_public,
                runtime_variant=endpoint_data.runtime_variant,
            )
        )

    async def list_errors(self, action: ListErrorsAction) -> ListErrorsActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        error_routes = (
            [r for r in endpoint_data.routings if r.status == RouteStatus.FAILED_TO_START]
            if endpoint_data.routings
            else []
        )

        return ListErrorsActionResult(
            error_info=[
                ErrorInfo(
                    session_id=route.error_data.get("session_id"), error=route.error_data["errors"]
                )
                for route in error_routes
            ],
            retries=endpoint_data.retries,
        )

    async def clear_error(self, action: ClearErrorAction) -> ClearErrorActionResult:
        # Clear errors with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            success = await self._admin_repository.clear_endpoint_errors_force(action.service_id)
        else:
            success = await self._repository.clear_endpoint_errors_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not success:
            raise ModelServiceNotFound

        return ClearErrorActionResult(success=True)

    async def update_route(self, action: UpdateRouteAction) -> UpdateRouteActionResult:
        # Update route traffic with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.update_route_traffic_force(
                self._valkey_live, action.route_id, action.service_id, action.traffic_ratio
            )
        else:
            endpoint_data = await self._repository.update_route_traffic_validated(
                self._valkey_live,
                action.route_id,
                action.service_id,
                action.traffic_ratio,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )
        if not endpoint_data:
            raise ModelServiceNotFound

        await self._agent_registry.notify_endpoint_route_update_to_appproxy(endpoint_data.id)

        return UpdateRouteActionResult(success=True)

    async def delete_route(self, action: DeleteRouteAction) -> DeleteRouteActionResult:
        # Get route with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            route_data = await self._admin_repository.get_route_by_id_force(
                action.route_id, action.service_id
            )
        else:
            route_data = await self._repository.get_route_by_id_validated(
                action.route_id,
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not route_data:
            raise RouteNotFound

        if route_data.status == RouteStatus.PROVISIONING:
            raise InvalidAPIParameters("Cannot remove route in PROVISIONING status")

        # Get session for destruction
        route_row = await self._repository.get_route_with_session(action.route_id)
        if not route_row:
            raise RouteNotFound

        await self._agent_registry.destroy_session(
            route_row.session_row,
            forced=False,
            reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
        )

        # Decrease endpoint replicas
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            await self._admin_repository.decrease_endpoint_replicas_force(action.service_id)
        else:
            await self._repository.decrease_endpoint_replicas_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        return DeleteRouteActionResult(success=True)

    async def generate_token(self, action: GenerateTokenAction) -> GenerateTokenActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        # Get scaling group info
        scaling_group_data = await self._repository.get_scaling_group_info(
            endpoint_data.resource_group
        )
        if not scaling_group_data:
            raise InvalidAPIParameters(f"Scaling group {endpoint_data.resource_group} not found")

        # Generate token via wsproxy
        body = {"user_uuid": str(endpoint_data.session_owner_id), "exp": action.expires_at}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{scaling_group_data.wsproxy_addr}/v2/endpoints/{endpoint_data.id}/token",
                json=body,
                headers={
                    "accept": "application/json",
                    "X-BackendAI-Token": scaling_group_data.wsproxy_api_token,
                },
            ) as resp:
                resp_json = await resp.json()
                if resp.status != HTTPStatus.OK:
                    raise EndpointNotFound(
                        f"Failed to generate token: {resp.status} {resp.reason} {resp_json}"
                    )
                token = resp_json["token"]

        # Create token in database
        token_id = uuid.uuid4()
        token_row = EndpointTokenRow(
            token_id,
            token,
            endpoint_data.id,
            endpoint_data.domain,
            endpoint_data.project,
            endpoint_data.session_owner_id,
        )

        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            token_data = await self._admin_repository.create_endpoint_token_force(token_row)
        else:
            token_data = await self._repository.create_endpoint_token_validated(
                token_row,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not token_data:
            raise ModelServiceNotFound

        # Convert data types
        service_token_data = ServiceEndpointTokenData(
            id=token_data.id,
            token=token_data.token,
            endpoint=token_data.endpoint,
            session_owner=token_data.session_owner,
            domain=token_data.domain,
            project=token_data.project,
            created_at=token_data.created_at,
        )
        return GenerateTokenActionResult(data=service_token_data)

    async def force_sync_with_app_proxy(self, action: ForceSyncAction) -> ForceSyncActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )
        if not endpoint_data:
            raise ModelServiceNotFound

        await self._agent_registry.notify_endpoint_route_update_to_appproxy(endpoint_data.id)

        return ForceSyncActionResult(success=True)

    async def modify_endpoint(self, action: ModifyEndpointAction) -> ModifyEndpointActionResult:
        result = await self._repository.modify_endpoint(
            action,
            self._agent_registry,
            self._config_provider.legacy_etcd_config_loader,
            self._storage_manager,
        )
        if action.modifier.replica_count_modified():
            # Notify appproxy to update routing info
            await self._deployment_controller.mark_lifecycle_needed(
                DeploymentLifecycleType.CHECK_REPLICA,
            )
        return ModifyEndpointActionResult(success=result.success, data=result.data)



from abc import ABC, abstractmethod
from collections.abc import Iterable

from ai.backend.manager.sokovan.scheduler.types import (
    SchedulingPredicate,
    SessionWorkload,
    SystemSnapshot,
)

from .exceptions import MultipleValidationErrors, SchedulingValidationError


class ValidatorRule(ABC):
    """
    An abstract base class for validator rules.
    Subclasses should implement the `validate` method to apply specific validation logic.
    """

    @abstractmethod
    def name(self) -> str:
        """Return the validator name for predicates."""
        raise NotImplementedError

    @abstractmethod
    def success_message(self) -> str:
        """Return a message describing successful validation."""
        raise NotImplementedError

    @abstractmethod
    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        """
        Validate a session workload against the system snapshot.

        Args:
            snapshot: The current system state snapshot
            workload: The session workload to validate

        Raises:
            SchedulingValidationError: If the workload fails validation
        """
        raise NotImplementedError


class SchedulingValidator:
    """
    A class that validates session workloads against a set of rules.
    It applies each rule to the provided workload to ensure it meets the required conditions.
    """

    _rules: Iterable[ValidatorRule]

    def __init__(self, rules: Iterable[ValidatorRule]) -> None:
        self._rules = rules

    def validate(
        self,
        snapshot: SystemSnapshot,
        workload: SessionWorkload,
        passed_phases: list[SchedulingPredicate],
        failed_phases: list[SchedulingPredicate],
    ) -> None:
        """
        Validate a session workload and update passed/failed phases lists.

        Args:
            snapshot: The current system state snapshot
            workload: The session workload to validate
            passed_phases: List to update with passed validation phases
            failed_phases: List to update with failed validation phases

        Raises:
            SchedulingValidationError: If any validation fails
        """
        errors: list[SchedulingValidationError] = []

        for rule in self._rules:
            try:
                rule.validate(snapshot, workload)
                # Validation passed - add to the passed list
                passed_phases.append(
                    SchedulingPredicate(name=rule.name(), msg=rule.success_message())
                )
            except SchedulingValidationError as e:
                # Validation failed - add to failed list and collect error
                failed_phases.append(SchedulingPredicate(name=rule.name(), msg=str(e)))
                errors.append(e)

        # If there were any failures, raise the appropriate exception
        if errors:
            if len(errors) == 1:
                raise errors[0]
            else:
                raise MultipleValidationErrors(errors)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from aiohttp import web

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class UpdatePasswordAction(AuthAction):
    request: web.Request
    user_id: uuid.UUID
    domain_name: str
    email: str
    old_password: str
    new_password: str
    new_password_confirm: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update_password"

    @property
    def hook_params(self) -> dict[str, str]:
        return {
            "old_password": self.old_password,
            "new_password": self.new_password,
            "new_password2": self.new_password_confirm,
        }


@dataclass
class UpdatePasswordActionResult(BaseActionResult):
    success: bool
    message: str

    @override
    def entity_id(self) -> Optional[str]:
        return None



from .create_channel import CreateChannelAction, CreateChannelActionResult
from .create_rule import CreateRuleAction, CreateRuleActionResult
from .delete_channel import DeleteChannelAction, DeleteChannelActionResult
from .delete_rule import DeleteRuleAction, DeleteRuleActionResult
from .get_channel import GetChannelAction, GetChannelActionResult
from .get_rule import GetRuleAction, GetRuleActionResult
from .list_channels import SearchChannelsAction, SearchChannelsActionResult
from .list_rules import SearchRulesAction, SearchRulesActionResult
from .process_notification import (
    ProcessedRuleSuccess,
    ProcessNotificationAction,
    ProcessNotificationActionResult,
)
from .update_channel import UpdateChannelAction, UpdateChannelActionResult
from .update_rule import UpdateRuleAction, UpdateRuleActionResult
from .validate_channel import ValidateChannelAction, ValidateChannelActionResult
from .validate_rule import ValidateRuleAction, ValidateRuleActionResult

__all__ = (
    "CreateChannelAction",
    "CreateChannelActionResult",
    "CreateRuleAction",
    "CreateRuleActionResult",
    "DeleteChannelAction",
    "DeleteChannelActionResult",
    "DeleteRuleAction",
    "DeleteRuleActionResult",
    "GetChannelAction",
    "GetChannelActionResult",
    "GetRuleAction",
    "GetRuleActionResult",
    "SearchChannelsAction",
    "SearchChannelsActionResult",
    "SearchRulesAction",
    "SearchRulesActionResult",
    "ProcessNotificationAction",
    "ProcessNotificationActionResult",
    "ProcessedRuleSuccess",
    "UpdateChannelAction",
    "UpdateChannelActionResult",
    "UpdateRuleAction",
    "UpdateRuleActionResult",
    "ValidateChannelAction",
    "ValidateChannelActionResult",
    "ValidateRuleAction",
    "ValidateRuleActionResult",
)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path

from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.logging.types import LogLevel

from .bootstrap import BootstrapComposer, BootstrapInput
from .components import ComponentsComposer, ComponentsInput, ComponentsResources
from .infrastructure import (
    InfrastructureComposer,
    InfrastructureInput,
    InfrastructureResources,
)


@dataclass
class DependencyInput:
    """Input required for complete dependency setup.

    Contains only the essential parameters: config file path and log level.
    """

    config_path: Path
    log_level: LogLevel = LogLevel.NOTSET


@dataclass
class DependencyResources:
    """Container for all initialized dependency resources.

    Holds all foundational dependencies in the correct initialization order:
    1. Infrastructure stage: valkey clients, database
    2. Components stage: storage manager, agent cache
    """

    infrastructure: InfrastructureResources
    components: ComponentsResources


class ManagerDependencyComposer(DependencyComposer[DependencyInput, DependencyResources]):
    """Composes all manager dependencies in the correct order.

    Composes the three-stage dependency initialization:
    1. Bootstrap: etcd and config provider
    2. Infrastructure: valkey clients and database
    3. Components: storage manager and agent cache
    """

    @property
    def stage_name(self) -> str:
        return "manager-dependencies"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: DependencyInput,
    ) -> AsyncIterator[DependencyResources]:
        """Compose all dependencies in order using the provided stack.

        Args:
            stack: The dependency stack to use for composition
            setup_input: Dependency input containing config path and log level

        Yields:
            DependencyResources containing all initialized resources
        """
        # Stage 1: Bootstrap (etcd + config provider)
        bootstrap_composer = BootstrapComposer()
        bootstrap_input = BootstrapInput(
            config_path=setup_input.config_path,
            log_level=setup_input.log_level,
        )
        bootstrap = await stack.enter_composer(
            bootstrap_composer,
            bootstrap_input,
        )

        # Stage 2: Infrastructure (valkey + database)
        infra_composer = InfrastructureComposer()
        infra_input = InfrastructureInput(
            config=bootstrap.config_provider.config,
            etcd=bootstrap.etcd,
        )
        infrastructure = await stack.enter_composer(
            infra_composer,
            infra_input,
        )

        # Stage 3: Components (storage manager + agent cache)
        components_composer = ComponentsComposer()
        components_input = ComponentsInput(
            config=bootstrap.config_provider.config,
            db=infrastructure.db,
            etcd=bootstrap.etcd,
        )
        components = await stack.enter_composer(
            components_composer,
            components_input,
        )

        # Yield all resources
        yield DependencyResources(
            infrastructure=infrastructure,
            components=components,
        )



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.object_storage.actions.create import (
    CreateObjectStorageAction,
    CreateObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.delete import (
    DeleteObjectStorageAction,
    DeleteObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.get import (
    GetObjectStorageAction,
    GetObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.get_download_presigned_url import (
    GetDownloadPresignedURLAction,
    GetDownloadPresignedURLActionResult,
)
from ai.backend.manager.services.object_storage.actions.get_upload_presigned_url import (
    GetUploadPresignedURLAction,
    GetUploadPresignedURLActionResult,
)
from ai.backend.manager.services.object_storage.actions.list import (
    ListObjectStorageAction,
    ListObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.update import (
    UpdateObjectStorageAction,
    UpdateObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.service import ObjectStorageService


class ObjectStorageProcessors(AbstractProcessorPackage):
    create: ActionProcessor[CreateObjectStorageAction, CreateObjectStorageActionResult]
    update: ActionProcessor[UpdateObjectStorageAction, UpdateObjectStorageActionResult]
    delete: ActionProcessor[DeleteObjectStorageAction, DeleteObjectStorageActionResult]
    get: ActionProcessor[GetObjectStorageAction, GetObjectStorageActionResult]
    list_storages: ActionProcessor[ListObjectStorageAction, ListObjectStorageActionResult]
    get_presigned_download_url: ActionProcessor[
        GetDownloadPresignedURLAction, GetDownloadPresignedURLActionResult
    ]
    get_presigned_upload_url: ActionProcessor[
        GetUploadPresignedURLAction, GetUploadPresignedURLActionResult
    ]

    def __init__(self, service: ObjectStorageService, action_monitors: list[ActionMonitor]) -> None:
        self.create = ActionProcessor(service.create, action_monitors)
        self.update = ActionProcessor(service.update, action_monitors)
        self.delete = ActionProcessor(service.delete, action_monitors)
        self.get = ActionProcessor(service.get, action_monitors)
        self.list_storages = ActionProcessor(service.list, action_monitors)
        self.get_presigned_download_url = ActionProcessor(
            service.get_presigned_download_url, action_monitors
        )
        self.get_presigned_upload_url = ActionProcessor(
            service.get_presigned_upload_url, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateObjectStorageAction.spec(),
            UpdateObjectStorageAction.spec(),
            DeleteObjectStorageAction.spec(),
            GetObjectStorageAction.spec(),
            ListObjectStorageAction.spec(),
            GetDownloadPresignedURLAction.spec(),
            GetUploadPresignedURLAction.spec(),
        ]



from __future__ import annotations

import logging
from decimal import Decimal
from typing import Mapping, Optional
from uuid import UUID

import trafaret as t

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.common.types import AccessKey, ResourceSlot
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.resource_preset.types import (
    ResourcePresetCreator,
    ResourcePresetModifier,
)

from .cache_source.cache_source import ResourcePresetCacheSource
from .db_source.db_source import ResourcePresetDBSource
from .types import CheckPresetsResult
from .utils import suppress_with_log

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


resource_preset_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.RESOURCE_PRESET_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ResourcePresetRepository:
    """Repository that orchestrates between DB and cache sources for resource preset operations."""

    _db_source: ResourcePresetDBSource
    _cache_source: ResourcePresetCacheSource
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_stat: ValkeyStatClient,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._db_source = ResourcePresetDBSource(db)
        self._cache_source = ResourcePresetCacheSource(valkey_stat)
        self._config_provider = config_provider

    @resource_preset_repository_resilience.apply()
    async def create_preset_validated(self, creator: ResourcePresetCreator) -> ResourcePresetData:
        """
        Creates a new resource preset.
        Raises ResourcePresetConflict if a preset with the same name and scaling group already exists.
        """
        preset = await self._db_source.create_preset(creator)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset creation"
        ):
            await self._cache_source.invalidate_all_presets()
        return preset

    @resource_preset_repository_resilience.apply()
    async def get_preset_by_id(self, preset_id: UUID) -> ResourcePresetData:
        """
        Gets a resource preset by ID.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        # Try cache first
        with suppress_with_log([Exception], message=f"Failed to get preset {preset_id} from cache"):
            preset = await self._cache_source.get_preset_by_id(preset_id)
            if preset:
                return preset

        # Fallback to DB
        preset = await self._db_source.get_preset_by_id(preset_id)
        with suppress_with_log([Exception], message=f"Failed to cache preset {preset_id}"):
            await self._cache_source.set_preset(preset)
        return preset

    @resource_preset_repository_resilience.apply()
    async def get_preset_by_name(self, name: str) -> ResourcePresetData:
        """
        Gets a resource preset by name.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        # Try cache first
        with suppress_with_log([Exception], message=f"Failed to get preset '{name}' from cache"):
            preset = await self._cache_source.get_preset_by_name(name)
            if preset:
                return preset

        # Fallback to DB
        preset = await self._db_source.get_preset_by_name(name)
        with suppress_with_log([Exception], message=f"Failed to cache preset '{name}'"):
            await self._cache_source.set_preset(preset)
        return preset

    @resource_preset_repository_resilience.apply()
    async def get_preset_by_id_or_name(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Gets a resource preset by ID or name.
        ID takes precedence if both are provided.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        return await self._db_source.get_preset_by_id_or_name(preset_id, name)

    @resource_preset_repository_resilience.apply()
    async def modify_preset_validated(
        self, preset_id: Optional[UUID], name: Optional[str], modifier: ResourcePresetModifier
    ) -> ResourcePresetData:
        """
        Modifies an existing resource preset.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        preset = await self._db_source.modify_preset(preset_id, name, modifier)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset modification"
        ):
            await self._cache_source.invalidate_preset(preset_id, name)
        return preset

    @resource_preset_repository_resilience.apply()
    async def delete_preset_validated(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Deletes a resource preset.
        Returns the deleted preset data.
        Raises ObjectNotFound if the preset doesn't exist.
        """
        preset = await self._db_source.delete_preset(preset_id, name)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset deletion"
        ):
            await self._cache_source.invalidate_preset(preset_id, name)
        return preset

    @resource_preset_repository_resilience.apply()
    async def list_presets(
        self, scaling_group_name: Optional[str] = None
    ) -> list[ResourcePresetData]:
        """
        Lists all resource presets.
        If scaling_group_name is provided, returns presets for that scaling group and global presets.
        """
        # Try cache first
        with suppress_with_log([Exception], message="Failed to get preset list from cache"):
            presets = await self._cache_source.get_preset_list(scaling_group_name)
            if presets is not None:
                return presets

        # Fallback to DB
        await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        presets = await self._db_source.list_presets(scaling_group_name)

        # Cache the result
        with suppress_with_log([Exception], message="Failed to cache preset list"):
            await self._cache_source.set_preset_list(presets, scaling_group_name)

        return presets

    @resource_preset_repository_resilience.apply()
    async def check_presets(
        self,
        access_key: AccessKey,
        user_id: UUID,
        group_name: str,
        domain_name: str,
        resource_policy: Mapping[str, str],
        scaling_group: Optional[str] = None,
    ) -> CheckPresetsResult:
        """
        Check resource presets availability and resource limits.
        """
        # Get configuration values
        known_slot_types = (
            await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        )
        # Try to get from cache first
        with suppress_with_log([Exception], message="Failed to get check presets data from cache"):
            cached_data = await self._cache_source.get_check_presets_data(
                access_key, group_name, domain_name, scaling_group
            )
            if cached_data:
                log.info(
                    "Cache hit for check_presets: {}, {}, {}", access_key, group_name, domain_name
                )
                return CheckPresetsResult.from_cache(cached_data)
        log.info(
            "Cache miss for check_presets, fetching from DB, {}, {}, {}",
            access_key,
            group_name,
            domain_name,
        )

        group_resource_visibility = await self._config_provider.legacy_etcd_config_loader.get_raw(
            "config/api/resources/group_resource_visibility"
        )
        group_resource_visibility = t.ToBool().check(group_resource_visibility)

        # Get all data from DB source
        db_data = await self._db_source.check_presets_data(
            access_key,
            user_id,
            group_name,
            domain_name,
            resource_policy,
            known_slot_types,
            scaling_group,
        )

        # Process the data and build response

        # Apply group resource visibility settings
        if not group_resource_visibility:
            nan_slots = ResourceSlot({
                str(k): Decimal("NaN") for k in db_data.known_slot_types.keys()
            })
            group_limits = nan_slots
            group_occupied = nan_slots
            group_remaining = nan_slots
        else:
            group_limits = db_data.keypair_data.group_limits
            group_occupied = db_data.keypair_data.group_occupied
            group_remaining = db_data.keypair_data.group_remaining

        result = CheckPresetsResult(
            presets=db_data.presets,
            keypair_limits=db_data.keypair_data.limits,
            keypair_using=db_data.keypair_data.occupied,
            keypair_remaining=db_data.keypair_data.remaining,
            group_limits=group_limits,
            group_using=group_occupied,
            group_remaining=group_remaining,
            scaling_group_remaining=db_data.keypair_data.scaling_group_remaining,
            scaling_groups=db_data.per_sgroup_data,
        )

        # Cache the result
        with suppress_with_log([Exception], message="Failed to cache check presets data"):
            cache_data = result.to_cache()
            await self._cache_source.set_check_presets_data(
                access_key,
                group_name,
                domain_name,
                scaling_group,
                cache_data,
            )

        return result



from typing import Any

from aiotools import aclosing
from etcd_client import WatchEventType

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.types import QueueSentinel
from ai.backend.manager.config.watchers.types import AbstractConfigWatcher


class EtcdConfigWatcher(AbstractConfigWatcher):
    _etcd: AsyncEtcd
    _config_prefix: str

    def __init__(self, etcd: AsyncEtcd, config_prefix: str = "ai/backend/config") -> None:
        self._etcd = etcd
        self._config_prefix = config_prefix

    async def watch(self) -> Any:
        async with aclosing(self._etcd.watch_prefix(self._config_prefix)) as agen:
            async for ev in agen:
                match ev:
                    case QueueSentinel.CLOSED | QueueSentinel.TIMEOUT:
                        break
                    case _:
                        if ev.event == WatchEventType.PUT:
                            yield ev
                        elif ev.event == WatchEventType.DELETE:
                            yield ev



"""keypairs_get_is_admin

Revision ID: 80176413d8aa
Revises: 4b8a66fb8d82
Create Date: 2017-09-14 16:01:59.994941

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql.expression import false

# revision identifiers, used by Alembic.
revision = "80176413d8aa"
down_revision = "4b8a66fb8d82"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "keypairs",
        sa.Column("is_admin", sa.Boolean(), nullable=False, default=False, server_default=false()),
    )
    op.create_index(op.f("ix_keypairs_is_admin"), "keypairs", ["is_admin"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_keypairs_is_admin"), table_name="keypairs")
    op.drop_column("keypairs", "is_admin")
    # ### end Alembic commands ###



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.manager.config.unified import ManagerUnifiedConfig
from ai.backend.manager.models.storage import StorageSessionManager

from .base import ComponentDependency


class StorageManagerDependency(ComponentDependency[StorageSessionManager]):
    """Provides StorageSessionManager lifecycle management."""

    @property
    def stage_name(self) -> str:
        return "storage-manager"

    @asynccontextmanager
    async def provide(
        self, setup_input: ManagerUnifiedConfig
    ) -> AsyncIterator[StorageSessionManager]:
        """Initialize and provide a storage session manager.

        Args:
            setup_input: Configuration containing volume settings

        Yields:
            Initialized StorageSessionManager
        """
        storage_manager = StorageSessionManager(setup_input.volumes)
        try:
            yield storage_manager
        finally:
            await storage_manager.aclose()

    def gen_health_checkers(self, resource: StorageSessionManager) -> None:
        """
        Return health checkers for storage manager.

        Currently no health checks are implemented for storage manager.

        Args:
            resource: The initialized storage session manager

        Returns:
            None (no health checks)
        """
        return None



import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass


class OperationStatus(enum.StrEnum):
    SUCCESS = "success"
    ERROR = "error"
    UNKNOWN = "unknown"
    RUNNING = "running"


@dataclass
class ActionSpec:
    entity_type: str
    operation_type: str

    def type(self) -> str:
        return f"{self.entity_type}:{self.operation_type}"


class AbstractProcessorPackage(ABC):
    @abstractmethod
    def supported_actions(self) -> list[ActionSpec]:
        """Get the list of action specs that this processors can handle."""
        raise NotImplementedError



"""Add `scanned_at`, `updated_at` columns to `artifacts`

Revision ID: 2cc8337d78a3
Revises: ba5207c55924
Create Date: 2025-09-03 10:52:39.346814

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2cc8337d78a3"
down_revision = "ba5207c55924"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "artifacts",
        sa.Column("scanned_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False),
    )
    op.add_column(
        "artifacts",
        sa.Column("updated_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("artifacts", "updated_at")
    op.drop_column("artifacts", "scanned_at")
    # ### end Alembic commands ###



from .context import RecorderContext
from .recorder import TransitionRecorder
from .types import ExecutionStep, PhaseDescriptor, StepDescriptor

__all__ = [
    "ExecutionStep",
    "PhaseDescriptor",
    "StepDescriptor",
    "TransitionRecorder",
    "RecorderContext",
]



import uuid

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.artifact_registry.db_source.db_source import (
    ArtifactRegistryDBSource,
)

artifact_registry_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.ARTIFACT_REGISTRY_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ArtifactRegistryRepository:
    """Repository layer that delegates to data source."""

    _db_source: ArtifactRegistryDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ArtifactRegistryDBSource(db)

    @artifact_registry_repository_resilience.apply()
    async def get_artifact_registry_data(self, registry_id: uuid.UUID) -> ArtifactRegistryData:
        return await self._db_source.get_artifact_registry_data(registry_id)

    @artifact_registry_repository_resilience.apply()
    async def get_artifact_registry_data_by_name(self, registry_name: str) -> ArtifactRegistryData:
        return await self._db_source.get_artifact_registry_data_by_name(registry_name)

    @artifact_registry_repository_resilience.apply()
    async def get_artifact_registry_datas(
        self, registry_ids: list[uuid.UUID]
    ) -> list[ArtifactRegistryData]:
        return await self._db_source.get_artifact_registry_datas(registry_ids)

    @artifact_registry_repository_resilience.apply()
    async def get_artifact_registry_type(self, registry_id: uuid.UUID) -> ArtifactRegistryType:
        return await self._db_source.get_artifact_registry_type(registry_id)

    @artifact_registry_repository_resilience.apply()
    async def list_artifact_registry_data(self) -> list[ArtifactRegistryData]:
        return await self._db_source.list_artifact_registry_data()
